/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.*;
import EDU.auburn.VGJ.graph.*;
import EDU.auburn.VGJ.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;
import EDU.utwente.CoCompose.CoCompose;

/**
 * A window class for editing and displaying CoCompose design models.
 * Based on EDU.auburn.VGJ.gui.GraphCanvas. Adapted for Swing.
 *
 * @author Dennis Wagelaar
 * @author Larry Barowski
 * @version 0.1.16, 31 October 2003
 * @see EDU.auburn.VGJ.gui.GraphCanvas
 */

public class ModelCanvas extends JPanel
  implements ErrorReporter, ModelUpdate
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  // selection flags
  private final static int NONE_   = 0;
  private final static int CENTER_ = 1;
  private final static int CORNER_ = 2;
  private final static int BOTTOM_ = 3;
  private final static int LEFT_   = 4;
  // maximum undo count
  private static final int MAX_UNDO = 20;
  // Mouse action values:
  /**
   * Mouse action value if no action currently in progress.
   */
  protected static final int MOUSE_NONE    = 0;
  /**
   * Mouse action value if a node is being created.
   */
  protected static final int MOUSE_NODE    = 1;
  /**
   * Mouse action value if an edge is being created.
   */
  protected static final int MOUSE_EDGE    = 2;
  /**
   * Mouse action value if a selection is being made.
   */
  protected static final int MOUSE_SELECT  = 3;
  /**
   * Mouse action value if a multiple selection is being made.
   */
  protected static final int MOUSE_MSELECT = 4;
  // Mouse mode values:
  /**
   * The select mode value.
   */
  public static final int SELECT_BOTH       = 1;
  /**
   * The create concepts mode value.
   */
  public static final int CREATE_EDGES      = 2;
  /**
   * The create concepts mode value.
   */
  public static final int CREATE_CONCEPTS   = 3;
  /**
   * The create solution roles mode value.
   */
  public static final int CREATE_SOLROLE    = 4;
  /**
   * The create composites mode value.
   */
  public static final int CREATE_COMPOSITES = 5;
  /**
   * The create composite roles mode value.
   */
  public static final int CREATE_ROLES      = 6;
  /**
   * The create published concepts mode value.
   */
  public static final int CREATE_PUBLISHED  = 7;
  /**
   * The create composites from repository mode value.
   */
  public static final int CREATE_REPOS      = 8;

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private double width_, height_; // drawable width and height
  private Dimension windowSize_ = new Dimension(-1, -1); // window viewable size
  // the scroll offsets - screen position + offset = physical position
  private double offsetx_ = 0, offsety_ = 0;
  private double minx_, miny_, maxx_, maxy_;
  private double scale_ = 1.0;
  private Matrix44 scaleMatrix_, shiftMatrix_;
  private Matrix44 rotxMatrix_, rotzMatrix_;
  private DPoint offset_ = new DPoint(0, 0);
  private Node newLinkNode_;
  private Node movingNode_;
  private Node selectedNode_;
  private double movingZ_, movingX_, movingY_;
  private double selectedRatio_;
  private Point selectedLink_;
  private Image backImage_ = null;
  private Color backColor_ = Color.white;
  private Color lineColor_ = Color.black;
  private int mouseMode_ = SELECT_BOTH;
  private int qualityCB_ = 1;
  private int quality_ = 1;
  private double hSpacing_ = 30, vSpacing_ = 40;
  private boolean xyPlane_ = true;
  private Model model_; // the model that is edited
  private Stack undoModels_ = new Stack();
  private Stack redoModels_ = new Stack();
  private Node drawingLinkNode_;
  private Frame frame_;
  private RepositoryContainer repository_;
  private EDU.utwente.CoCompose.lang.Composite newReposComposite_;
  private boolean hoverSelect_;
  private DPoint3 lastLinkPoint_ = null;
  private int pathLength_, pathArraySize_;
  private DPoint3[] pathArray_;
  private int newLinkX_, newLinkY_;
  private int selected_ = NONE_;
  private double moveX_, moveY_;
  private int multiSelectX_, multiSelectY_, multiSelectX2_, multiSelectY2_;
  private int origMode_;
  private int enabled_ = 0; // 0 = true (multi-level)
  private boolean dragStarted_ = false; // flag to determine when to update undo status
  /**
   * The current mouse action being performed
   */
  protected int currentMouseAction = MOUSE_NONE;
  /**
   * The coordinate transformation matrix to transform from graph coordinates
   * to viewing coordinates.
   */
  protected Matrix44 viewTransform;
  /**
   * The coordinate transformation matrix to transform from viewing coordinates
   * to graph coordinates.
   */
  protected Matrix44 moveTransform;
  /**
   * Only one <code>ChangeEvent</code> is needed per instance since the
   * event's only state is the source property.  The source of events
   * generated is always "this".
   */
  protected transient ChangeEvent changeEvent;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a ModelCanvas.
   *
   * @param model_in The Model object that will be edited.
   * @param frame_in The parent frame that contains this object.
   */
  public ModelCanvas(Model model_in, Frame frame_in)
  {
    super();
    model_ = model_in;
    frame_ = frame_in;
    try
    {
      jbInit();
      mInit();
    }
    catch(Exception e)
    {
      e.printStackTrace();
    }
  } //ModelCanvas(Model model_in, Frame frame_in)

  /**
   * Constructs a ModelCanvas.
   *
   * @param frame_in The parent frame that contains this object.
   */
  public ModelCanvas(Frame frame_in)
  {
    this(new Model(), frame_in);
  } //ModelCanvas(Frame frame_in)

  /**
   * Constructs a ModelCanvas.
   * This constructor is only for the JBuilder Designer, do not use it!
   */
  public ModelCanvas()
  {
    this(new Model(), null);
  } //ModelCanvas()

  //////////////////////////////////////////////////////////////////////////////
  // Private methods
  //////////////////////////////////////////////////////////////////////////////

  /* JBuilder Component initialization. */
  private void jbInit() throws Exception
  {
    this.setPreferredSize(new Dimension(400, 400));
    this.addMouseMotionListener(new java.awt.event.MouseMotionAdapter()
    {
      public void mouseDragged(MouseEvent e)
      {
        this_mouseDragged(e);
      }
      public void mouseMoved(MouseEvent e)
      {
        this_mouseMoved(e);
      }
    });
    this.addMouseListener(new java.awt.event.MouseAdapter()
    {
      public void mousePressed(MouseEvent e)
      {
        this_mousePressed(e);
      }
      public void mouseReleased(MouseEvent e)
      {
        this_mouseReleased(e);
      }
      public void mouseExited(MouseEvent e)
      {
        this_mouseExited(e);
      }
    });
  } //jbInit

  /* Manual initialization. */
  private void mInit()
  {
    computeBounds();
    // initialize matrices
    scaleMatrix_ = new Matrix44();
    scaleMatrix_.matrix[0][0] = scaleMatrix_.matrix[1][1] =
      scaleMatrix_.matrix[2][2] = scaleMatrix_.matrix[3][3] = 1.0;
    shiftMatrix_ = new Matrix44(scaleMatrix_);
    rotxMatrix_ = new Matrix44(scaleMatrix_);
    rotxMatrix_.matrix[1][1] = rotxMatrix_.matrix[2][2] = -1.0;
    rotxMatrix_.matrix[2][1] = -(rotxMatrix_.matrix[1][2] = 0.0);
    rotzMatrix_ = new Matrix44(scaleMatrix_);
    updateViewTransform();
  } //mInit

  /* Computes canvas bounds. */
  private void computeBounds()
  {
    double oldminx = minx_;
    double oldminy = miny_;

    minx_ = miny_ = maxx_ = maxy_ = 0;

    Node tmpnode = model_.firstNode();
    DPoint tmppoint;
    DDimension tmpdim;

    if (tmpnode != null)
    {
      tmppoint = tmpnode.getPosition();
      tmpdim = tmpnode.getBoundingBox();

      minx_ = tmppoint.x - tmpdim.width / 2.0;
      maxx_ = tmppoint.x + tmpdim.width / 2.0;
      miny_ = tmppoint.y - tmpdim.height / 2.0;
      maxy_ = tmppoint.y + tmpdim.height / 2.0;

      tmpnode = model_.nextNode(tmpnode);
      while (tmpnode != null)
      {
        tmppoint = tmpnode.getPosition();
        tmpdim = tmpnode.getBoundingBox();

        double w = tmpdim.width / 2.0;
        double h = tmpdim.height / 2.0;

        if (tmppoint.x - w < minx_)
          minx_ = tmppoint.x - w;
        if (tmppoint.x + w > maxx_)
          maxx_ = tmppoint.x + w;
        if (tmppoint.y - h < miny_)
          miny_ = tmppoint.y - h;
        if (tmppoint.y + h > maxy_)
          maxy_ = tmppoint.y + h;
        tmpnode = model_.nextNode(tmpnode);
      }
    }
    // 3D approach.
    double maxdim = Math.abs(maxy_);
    if (Math.abs(miny_) > maxdim)
      maxdim = Math.abs(miny_);
    if (Math.abs(minx_) > maxdim)
      maxdim = Math.abs(minx_);
    if (Math.abs(maxx_) > maxdim)
      maxdim = Math.abs(maxx_);

    maxx_ = maxy_ = maxdim;
    minx_ = miny_ = -maxdim;
    width_ = height_ = 2.0 * maxdim;
  } //computeBounds

  /* Update viewTransform and moveTransform matrix. */
  private void updateViewTransform()
  {
    viewTransform = new Matrix44(shiftMatrix_);
    viewTransform.mult(scaleMatrix_);
    viewTransform.mult(rotxMatrix_);
    viewTransform.mult(rotzMatrix_);

    moveTransform = new Matrix44(rotzMatrix_);
    moveTransform.matrix[1][0] = -moveTransform.matrix[1][0];
    moveTransform.matrix[0][1] = -moveTransform.matrix[0][1];

    Matrix44 tmp_matrix = new Matrix44(rotxMatrix_);
    tmp_matrix.matrix[2][1] = -tmp_matrix.matrix[2][1];
    tmp_matrix.matrix[1][2] = -tmp_matrix.matrix[1][2];
    moveTransform.mult(tmp_matrix);

    tmp_matrix.setTo(scaleMatrix_);
    tmp_matrix.matrix[0][0] = 1.0 / tmp_matrix.matrix[0][0];
    tmp_matrix.matrix[1][1] = 1.0 / tmp_matrix.matrix[1][1];
    tmp_matrix.matrix[2][2] = 1.0 / tmp_matrix.matrix[2][2];
    moveTransform.mult(tmp_matrix);

    tmp_matrix.setTo(shiftMatrix_);
    tmp_matrix.matrix[0][3] = - tmp_matrix.matrix[0][3];
    tmp_matrix.matrix[1][3] = - tmp_matrix.matrix[1][3];
    moveTransform.mult(tmp_matrix);

    viewTransform.scale = scale_;
  } //updateViewTransform

  /* Computes drawing offset. */
  private void computeDrawOffset()
  {
    setOffset((-offsetx_ - minx_) * scale_ + windowSize_.width,
              (-offsety_ - miny_) * scale_ + windowSize_.height);
  } //computeDrawOffset

  /* Sets the drawing offset. */
  private void setOffset(double offx, double offy)
  {
    offset_.x = offx;
    offset_.y = offy;
    shiftMatrix_.matrix[0][3] = offx;
    shiftMatrix_.matrix[1][3] = offy;
    updateViewTransform();
  } //setOffset

  /* Creates a CompositeRole at (x_in, y_in) if legal. Throws an exception if
   * CompositeRole is not legal. */
  private void createRole(int x_in, int y_in) throws CoComposeException
  {
    EDU.utwente.CoCompose.lang.Composite thisComposite;

    if ((thisComposite = findNearestComposite(x_in, y_in)) != null)
    {
      CompositeRole thisRole = new CompositeRole();
      // role ids unique within composites
      model_.setInsertNode(thisRole);
      thisComposite.addPartConcept(thisRole);
    }
    else
    {
      throw new CoComposeException("Composite roles can only be created inside composites");
    }
  } //createRole

  /* Creates a PublishedConcept at (x_in, y_in) if legal. Throws an exception if
   * PublishedConcept is not legal. */
  private void createPublished(int x_in, int y_in) throws CoComposeException
  {
    EDU.utwente.CoCompose.lang.Composite thisComposite;

    if ((thisComposite = findNearestComposite(x_in, y_in)) != null)
    {
      PublishedConcept thisPublished = new PublishedConcept();
      // role ids unique within composites
      model_.setInsertNode(thisPublished);
      thisComposite.addPartConcept(thisPublished);
    }
    else
    {
      throw new CoComposeException("Published concepts can only be created inside composites");
    }
  } //createPublished

  /* Creates a Composite from the repository if possible. */
  private void createRepositoryComposite() throws CoComposeException
  {
    // repository may not be null
    if (repository_ == null)
    {
      throw new CoComposeException("Repository not set");
    }
    // repository must return a composite
    newReposComposite_ = repository_.getSelectedComposite();
    if (newReposComposite_ == null)
    {
      throw new CoComposeException("No composite type selected from repository");
    }
    // insert composite and roles
    model_.setInsertNode(newReposComposite_);
  } //createRepositoryComposite

  /* Handles a mouse click when currentMouseAction = MOUSE_NONE. */
  private void handleMouseClickNone(MouseEvent e)
  {
    int x_in = e.getX();
    int y_in = e.getY();
    boolean shift = (e.getModifiers() & MouseEvent.SHIFT_MASK) != 0;
    // Determine which action to take
    int mode = getMouseMode();
    switch (mode)
    {
      case SELECT_BOTH:  // selection mode
        currentMouseAction = MOUSE_SELECT;
        select(x_in, y_in, shift);
        break;
      case CREATE_EDGES: // create edges mode
        currentMouseAction = MOUSE_EDGE;
        insertLink(x_in, y_in);
        break;
      default:           // other mode
        currentMouseAction = MOUSE_NODE;
        insertNode(x_in, y_in);
        break;
    }
    // Get transformed position
    DPoint3 pos = new DPoint3(x_in, y_in, 0.0);
    pos.transform(moveTransform);
    // set new label
    String string = getLabel(pos.x, pos.y, pos.z, true);
    firePropertyChange("Label", "", string);
  } //handleMouseClickNone

  /* Handles a mouse right click when currentMouseAction = MOUSE_NONE. */
  private void handleMouseRightClickNone(MouseEvent e)
  {
    int x_in = e.getX();
    int y_in = e.getY();
    boolean shift = (e.getModifiers() & MouseEvent.SHIFT_MASK) != 0;
    // Always do select mode
    currentMouseAction = MOUSE_SELECT;
    select(x_in, y_in, shift);
    // Get transformed position
    DPoint3 pos = new DPoint3(x_in, y_in, 0.0);
    pos.transform(moveTransform);
    // set new label
    String string = getLabel(pos.x, pos.y, pos.z, true);
    firePropertyChange("Label", "", string);
  } //handleMouseClickNone

  /* Handles a mouse double click when currentMouseAction = MOUSE_NONE. */
  private void handleMouseDoubleClickNone(MouseEvent e)
  {
    currentMouseAction = MOUSE_NONE;
    setNodeProperties();
  } //handleMouseClickNone

  /* Handles a mouse click when currentMouseAction = MOUSE_EDGE. */
  private void handleMouseClickLink(MouseEvent e)
  {
    int x_in = e.getX();
    int y_in = e.getY();
    if (newLinkNode_ != null)
    {
      Node tmpnode = findNearestNode(x_in, y_in);
      if (tmpnode != null)
      {  // Finish edge.
        if (pathLength_ == 0 && newLinkNode_ != tmpnode)
        {
          try
          {
            model_.insertLink(new Link(newLinkNode_, tmpnode, new DPoint3[]{}));
          }
          catch (CoComposeException exc)
          {
            reportError(exc);
          }
        }
        else
        {
          DPoint3[] tmp_array;
          if (pathLength_ == 0)
          { // Must be self-edge with no intermediate points
            // Add some intermediate points.
            tmp_array = new DPoint3[2];
            DPoint3 pos = tmpnode.getPosition3();
            pos.transform(viewTransform);
            DPoint3 pos2 = new DPoint3(pos);
            DDimension3 size = tmpnode.getBoundingBox3();
            double w = size.width / 2.0 * scale_;
            if (w < 10)
              w = 10;
            double h = size.height / 2.0 * scale_ + w * 1.5;
            pos.translate(-w, -h, 0);
            pos2.translate(w, -h, 0);
            pos.transform(moveTransform);
            pos2.transform(moveTransform);
            tmp_array[0] = clip(pos);
            tmp_array[1] = clip(pos2);
          }
          else
          {
            tmp_array = new DPoint3[pathLength_];
            System.arraycopy(pathArray_, 0, tmp_array, 0, pathLength_);
          }
          try
          {
            model_.insertLink(new Link(newLinkNode_, tmpnode, tmp_array));
          }
          catch (CoComposeException exc)
          {
            reportError(exc);
          }
        }
        newLinkNode_ = null;
        currentMouseAction = MOUSE_NONE;
        repaint();
      }
      else  // New edge path point.
      {
        DPoint3 pos = new DPoint3(x_in, y_in, 0.0);
        pos.transform(moveTransform);
        clip(pos);

        DPoint3 p2;
        if (lastLinkPoint_ != null)
          p2 = new DPoint3(lastLinkPoint_);
        else
          p2 = newLinkNode_.intersectWithLineTo(pos, xyPlane_, quality_);

        p2.transform(viewTransform);

        Graphics back = backImage_.getGraphics();
        back.setColor(Link.DEFAULT_EDGE_COLOR);
        back.drawLine((int) x_in, (int) y_in, (int) p2.x, (int) p2.y);
        back.dispose();

        lastLinkPoint_ = pos;

        if (pathLength_ >= pathArraySize_)
        {
          pathArraySize_ = pathLength_ * 2;
          DPoint3[] new_array = new DPoint3[pathArraySize_];
          System.arraycopy(pathArray_, 0, new_array, 0, pathLength_);
          pathArray_ = new_array;
        }

        pathArray_[pathLength_] = lastLinkPoint_;
        pathLength_++;
      }
    }
  } //handleMouseClickLink

  /* Handles a mouse release when currentMouseAction = MOUSE_NODE. */
  private void handleMouseReleaseNode(MouseEvent e)
  {
    // reset flag
    dragStarted_ = false;
    // if just inserted a composite from repository, retore the roles offset
    if (newReposComposite_ != null)
    {
      newReposComposite_.restoreOffsets();
      newReposComposite_ = null;
    }
    currentMouseAction = MOUSE_NONE;
    // new bound cause repaint and re-adjustment of observers.
    computeBounds();
    repaint();
    fireStateChanged();
  } //handleMouseReleaseNode

  /* Handles a mouse release when currentMouseAction = MOUSE_SELECT. */
  private void handleMouseReleaseSelect(MouseEvent e)
  {
    currentMouseAction = MOUSE_NONE;
    repaint();
    fireStateChanged();
  } //handleMouseReleaseSelect

  /* Handles a mouse release when currentMouseAction = MOUSE_MSELECT. */
  private void handleMouseReleaseMSelect(MouseEvent e)
  {
    currentMouseAction = MOUSE_NONE;
    multiSelect();
  } //handleMouseReleaseSelect

  /* Handles a mouse drag event when currentMouseAction = MOUSE_NODE. */
  private void handleMouseDragNode(MouseEvent e)
  {
    if (!dragStarted_)
    { // update undo state and set flag
      updateUndo();
      dragStarted_ = true;
    }
    // Get transformed position
    DPoint3 pos = new DPoint3(e.getX(), e.getY(), 0.0);
    pos.transform(moveTransform);
    movingNode_.setPosition(pos);
    // if just inserted a composite from repository, retore the roles offset
    if (newReposComposite_ != null)
    {
      newReposComposite_.restoreOffsets();
    }
    // Redraw
    setWireframe(true);
    repaint();
  } //handleMouseDragNode

  /* Handles a mouse drag event when currentMouseAction = MOUSE_EDGE. */
  private void handleMouseDragLink(MouseEvent e)
  {
    if (newLinkNode_ != null)
    {
      newLinkX_ = e.getX();
      newLinkY_ = e.getY();
      // draw
      repaint();
    }
  } //handleMouseDragLink

  /* Handles a mouse drag event when currentMouseAction = MOUSE_SELECT. */
  private void handleMouseDragSelect(MouseEvent e)
  {
    if (selectedNode_ == null)
      return;

    int x_in = e.getX();
    int y_in = e.getY();
    // If a node is dragged far enough, start moving it.
    if (selected_ == NONE_)
    {
      DPoint3 tmppos = selectedNode_.getPosition3();
      tmppos.transform(viewTransform);

      moveX_ = tmppos.x;
      moveY_ = tmppos.y;
      int xoffs = x_in - (int)moveX_;
      int yoffs = y_in - (int)moveY_;
      if (xoffs * xoffs + yoffs * yoffs > 9) // Three pixels
      {
        selected_ = CENTER_;

        movingZ_ = tmppos.z;
        movingX_ = tmppos.x;
        movingY_ = tmppos.y;

        for (Node tmpnode = model_.firstNode(); tmpnode != null; tmpnode = model_.nextNode(tmpnode))
        {
          tmpnode.saveState();
        }
        Enumeration edges = model_.getLinks();
        while (edges.hasMoreElements())
        {
          ((Link) (edges.nextElement())).saveState();
        }
      }
    }

    if (selected_ == CENTER_)
    {
      int xoffs = Math.round((x_in - (int) moveX_) / 10) * 10;
      int yoffs = Math.round((y_in - (int) moveY_) / 10) * 10;

      for (Node tmpnode = model_.firstNode(); tmpnode != null; tmpnode = model_.nextNode(tmpnode))
      {
        if (tmpnode.getSelected())
          tmpnode.slide(moveTransform, viewTransform, xoffs, yoffs);
      }

      Enumeration edges = model_.getLinks();
      while(edges.hasMoreElements())
      {
        Link edge = (Link) (edges.nextElement());
        if (edge.tail().getSelected() && edge.head().getSelected())
          edge.slide(moveTransform, viewTransform, xoffs, yoffs);
      }
    }
    else if (selected_ != NONE_)
    {
      double ratiox = 1.0, ratioy = 1.0, ratioz = 1.0;
      if (selected_ == CORNER_)
        ratiox = ratioy = ratioz = Math.max(
          Math.abs(x_in - movingX_) / Math.abs(moveX_ - movingX_),
          Math.abs(y_in - movingY_) / Math.abs(moveY_ - movingY_));
      else if (selected_ == BOTTOM_)
        ratioy = Math.abs(y_in - movingY_) / Math.abs(moveY_ - movingY_);
      else if (selected_ == LEFT_)
        ratiox = Math.abs(x_in - movingX_) / Math.abs(moveX_ - movingX_);
      for (Node tmpnode = model_.firstNode(); tmpnode != null; tmpnode = model_.nextNode(tmpnode))
      {
        if (tmpnode.getSelected())
          tmpnode.scale(ratiox, ratioy, ratioz);
      }
    }
    // repaint
    setWireframe(true);
    repaint();
  } //handleMouseDragSelect

  /* Handles a mouse drag event when currentMouseAction = MOUSE_MSELECT. */
  private void handleMouseDragMSelect(MouseEvent e)
  {
    multiSelectX2_ = e.getX();
    multiSelectY2_ = e.getY();
    Graphics g = getGraphics();
    paintComponentBuffer(g);
    g.dispose();
  } //handleMouseDragMSelect

  /* Sets link with given endpoints to selected or non-selected. */
  private void setLinkSelected_(int n1, int n2, boolean state)
  {
    Edge link = model_.getEdge(Math.min(n1, n2), Math.max(n1, n2));
    link.selected = state;
  } //setLinkSelected

  /* Formats the given number as PostScript representation. */
  private String PSnum_(double num)
  {
    if (num > 0.0)
      return String.valueOf(num) + " ";
    else if (num < 0.0)
      return String.valueOf(-num) + " neg ";
    else
      return "0 ";
  } //PSnum_

  /* Mouse button pressed. */
  void this_mousePressed(MouseEvent e)
  {
    // don't handle events when disabled
    if (!getEnabled())
      return;
    // mouse button 1 clicked
    boolean button1 = ((e.getModifiers() & MouseEvent.BUTTON1_MASK) != 0);
    // mouse button 2 clicked
    boolean button2 = ((e.getModifiers() & MouseEvent.BUTTON2_MASK) != 0);
    // double clicked
    boolean doubleclick = button1 && (e.getClickCount() == 2);
    // original mouse mode
    origMode_ = getMouseMode();

    // handle 1st button double click
    if (doubleclick)
    {
      dispatchMouseDoubleClickAction(e);
    }
    // handle 1st button click
    else if (button1)
    {
      try
      {
        dispatchNodeCreation(e);
        // if temporary select mode needed, set new mode
        if (hoverSelect_)
          setMouseMode(SELECT_BOTH);
        // handle mouse click
        dispatchMouseClickAction(e);
      }
      catch (CoComposeException exc)
      {
        reportError(exc);
      }
    }
    // handle 2nd button click
    else
    {
      dispatchMouseRightClick(e);
    }
  } //this_mousePressed

  /* Mouse button released. */
  void this_mouseReleased(MouseEvent e)
  {
    // don't handle events when disabled
    if (!getEnabled())
      return;
    // handle any button release
    dispatchMouseReleaseAction(e);
    // if temporary select mode was needed, reset original mode
    if (hoverSelect_)
      setMouseMode(origMode_);
  } //this_mouseReleased

  /* Mouse dragged. */
  void this_mouseDragged(MouseEvent e)
  {
    // handle dragging
    dispatchMouseDragAction(e);
  } //this_mouseDragged

  /* Mouse moved. */
  void this_mouseMoved(MouseEvent e)
  {
    // same behaviour as mouse dragged
    this_mouseDragged(e);
  } //this_mouseMoved

  /* Mouse exited this canvas. */
  void this_mouseExited(MouseEvent e)
  {
    // set new label
    String string = getLabel(0, 0, 0, false);
    firePropertyChange("Label", "", string);
  } //this_mouseExited

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets whether or not this canvas is enabled. A canvas that is enabled
   * may respond to user input, while a canvas that is not enabled cannot
   * respond to user input.
   *
   * @return True if canvas is enabled, otherwise false.
   */
  protected boolean getEnabled()
  {
    return (enabled_ == 0);
  } //getEnabled

  /**
   * Paints the canvas elements on top of the panel.
   *
   * @param g The graphics context.
   */
  protected synchronized void paintComponent(Graphics g)
  {
    // if resized, discard backbuffer
    Dimension winsize = getSize();
    if (winsize.width != windowSize_.width || winsize.height != windowSize_.height)
    {
      // Maintain the center point.
      if (windowSize_.width > 0) // not first time
      {
        setOffset(offset_.x + (winsize.width - windowSize_.width) / 2,
                  offset_.y + (winsize.height - windowSize_.height) / 2);
      }
      else // initialize to centered
      {
        setOffset(0.5 * (winsize.width - (minx_ + maxx_) * scale_),
                  0.5 * (winsize.height - (miny_ + maxy_) * scale_));
      }
      windowSize_.width = winsize.width;
      windowSize_.height = winsize.height;
      backImage_ = null;
      // post an event indicating a size change
      fireStateChanged();
    }
    // create new backbuffer if needed
    if (backImage_ == null)
    {
      backImage_ = createImage(windowSize_.width, windowSize_.height);
    }
    if (currentMouseAction != MOUSE_EDGE)
    { // do not repaint buffer when creating an edge
      // get backbuffer graphics context
      Graphics back = backImage_.getGraphics();
      // draw backbuffer
      try
      {
        back.setColor(getBackColor());
        back.fillRect(0, 0, windowSize_.width, windowSize_.height);
        back.setColor(getLineColor());
        back.setPaintMode();
        drawAxes(back);
        drawObjects(false, back);
      }
      finally
      {
        back.dispose();
      }
    }
    // draw screen
    paintComponentBuffer(g);
  } //paintComponent

  /**
   * Paints the viewport elements on top of the panel. Does a partial paint
   * by using the backbuffer.
   *
   * @param g The graphics context.
   */
  protected synchronized void paintComponentBuffer(Graphics g)
  {
    Graphics graphics = getComponentGraphics(g);
    try
    {
      graphics.drawImage(backImage_, 0, 0, null);
      // Draw selected objects directly to screen.
      drawObjects(true, graphics);
    }
    finally
    {
      graphics.dispose();
    }
  } //paintComponentBuffer

  /**
   * Notifies all listeners that have registered interest for
   * notification on this event type.  The event instance
   * is lazily created using the parameters passed into
   * the fire method.
   *
   * @see EventListenerList
   */
  protected void fireStateChanged()
  {
    // Guaranteed to return a non-null array
    Object[] listeners = listenerList.getListenerList();
    // Process the listeners last to first, notifying
    // those that are interested in this event
    for (int i = listeners.length - 2; i >= 0; i -= 2)
    {
      if (listeners[i] == ChangeListener.class)
      {
        // Lazily create the event:
        if (changeEvent == null)
          changeEvent = new ChangeEvent(this);
        ((ChangeListener) listeners[i + 1]).stateChanged(changeEvent);
      }
    }
  } //fireStateChanged

  /**
   * Gets the label text, e.g. for printing on a panel.
   *
   * @param x The X coordinate to print.
   * @param y The X coordinate to print.
   * @param z The X coordinate to print.
   * @param mousein If true, the mouse is over this canvas.
   * @return The label text.
   */
  protected synchronized String getLabel(double x, double y, double z, boolean mousein)
  {
    String string = " ";
    // mouse position
    if (mousein)
      string = string.concat("x: " + (int) x + "  y: " + (int) y + "  z: " + (int) z);
    // selected node position
    if (selectedNode_ != null)
    {
      DPoint3 pos = selectedNode_.getPosition3();
      DDimension3 bbox = selectedNode_.getBoundingBox3();
      string = string.concat("   " + selectedNode_.getLabel() +
                             "   x: " + (int) pos.x + "  y: " + (int) pos.y + "  z: " + (int) pos.z +
                             "   w: " + (int) bbox.width + "  h: " + (int) bbox.height + "  d: " + (int) bbox.depth);
    }
    // selected edge position
    if (selectedLink_ != null)
    {
      string = string.concat("   Link (" + selectedLink_.x + "," +
                             selectedLink_.y + ")");
    }
    return string;
  } //getLabel

  /**
   * Draw selected or unselected objects.
   *
   * @param selected If true, draws selected elements, otherwise unselected.
   * @param graphics The graphics context
   */
  protected void drawObjects(boolean selected, Graphics graphics)
  {
    // Draw nodes.
    Node tmpnode;
    for (tmpnode = model_.firstNode(); tmpnode != null; tmpnode = model_.nextNode(tmpnode))
    {
      if (tmpnode.getSelected() == selected)
        tmpnode.draw(this, graphics, viewTransform, quality_);
    }
    // Draw edges.
    Enumeration links = model_.getLinks();
    while (links.hasMoreElements())
    {
      Link link = (Link) (links.nextElement());
      Node head = link.head(), tail = link.tail();
      boolean sel = link.selected;
      if ((link.selected == selected) ||
          (head.getSelected() == selected) || (tail.getSelected() == selected))
      {
        link.draw(graphics, viewTransform, xyPlane_, quality_);
      }
    }
    // draw any edge under construction
    if (newLinkNode_ != null && selected)
    {
      // Get transformed position
      DPoint3 pos = new DPoint3((double) newLinkX_, (double) newLinkY_, 0.0);
      pos.transform(moveTransform);
      DPoint3 p2;
      if (lastLinkPoint_ != null)
        p2 = new DPoint3(lastLinkPoint_);
      else
        p2 = newLinkNode_.intersectWithLineTo(pos, xyPlane_, quality_);
      p2.transform(viewTransform);
      graphics.setColor(Link.DEFAULT_EDGE_COLOR);
      graphics.drawLine(newLinkX_, newLinkY_, (int) p2.x, (int) p2.y);
    }
    // draw a selection box if multi-selecting
    if (currentMouseAction == MOUSE_MSELECT)
    {
      int sx = Math.min(multiSelectX_, multiSelectX2_);
      int sy = Math.min(multiSelectY_, multiSelectY2_);
      int sw = Math.abs(multiSelectX_ - multiSelectX2_);
      int sh = Math.abs(multiSelectY_ - multiSelectY2_);
      if (multiSelectX2_ != -1)
      {
        graphics.setColor(getLineColor());
        graphics.drawRect(sx, sy, sw, sh);
      }
    }
  } //drawObjects

  /**
   * Draws the coordinate axes.
   *
   * @param graphics The graphics context
   */
  protected void drawAxes(Graphics graphics)
  {
    String letter;
    double lx, ly;
    FontMetrics fm = graphics.getFontMetrics();

    graphics.setColor(getLineColor());

    DPoint3 p2 = new DPoint3();
    p2.move(25, 0, 0);
    p2.transform(rotzMatrix_);
    p2.transform(rotxMatrix_);
    graphics.drawLine(40, 40, 40 + (int)p2.x, 40 + (int)p2.y);

    p2.move(32, 0, 0);
    p2.transform(rotzMatrix_);
    p2.transform(rotxMatrix_);
    letter = new String("X");
    lx = p2.x - fm.stringWidth(letter) / 2.0;
    ly = p2.y + fm.getAscent() / 2.0;
    graphics.drawString(letter, 40 + (int)lx, 40 + (int)ly);

    p2.move(0, 25, 0);
    p2.transform(rotzMatrix_);
    p2.transform(rotxMatrix_);
    graphics.drawLine(40, 40, 40 + (int)p2.x, 40 + (int)p2.y);

    p2.move(0, 32, 0);
    p2.transform(rotzMatrix_);
    p2.transform(rotxMatrix_);
    letter = new String("Y");
    lx = p2.x - fm.stringWidth(letter) / 2.0;
    ly = p2.y + fm.getAscent() / 2.0;
    graphics.drawString(letter, 40 + (int)lx, 40 + (int)ly);

    p2.move(0, 0, 25);
    p2.transform(rotzMatrix_);
    p2.transform(rotxMatrix_);
    graphics.drawLine(40, 40, 40 + (int)p2.x, 40 + (int)p2.y);

    p2.move(0, 0, 32);
    p2.transform(rotzMatrix_);
    p2.transform(rotxMatrix_);
    letter = new String("Z");
    lx = p2.x - fm.stringWidth(letter) / 2.0;
    ly = p2.y + fm.getAscent() / 2.0;
    graphics.drawString(letter, 40 + (int)lx, 40 + (int)ly);
  } //drawAxes

  /**
   * Finds the node that lies nearest to the given coordinates, if any.
   *
   * @param x The X mouse coordinate.
   * @param y The Y mouse coordinate.
   * @return The nearest node or null.
   */
  protected Node findNearestNode(double x, double y)
  {
    Node node = null;

    Node tmpnode;
    DPoint3 pos;
    DDimension bbox;

    double closest = (width_ + height_) * scale_;

    for (tmpnode = model_.firstNode(); tmpnode != null; tmpnode = model_.nextNode(tmpnode))
    {
      if (!tmpnode.isVisible())
        continue;
      pos = tmpnode.getPosition3();
      bbox = tmpnode.getBoundingBox();

      pos.transform(viewTransform);

      pos.x = Math.abs(pos.x - x);
      pos.y = Math.abs(pos.y - y);

      if (pos.x < bbox.width / 2.0 * scale_ + 1 && pos.y < bbox.height / 2.0 * scale_ + 1)
      {
        if (pos.x + pos.y <= closest)  // rough estimate of closeness
        {
          closest = pos.x + pos.y;
          node = tmpnode;
        }
      }
    }
    return node;
  } //findNearestNode

  /**
   * Finds the edge that lies nearest to the given coordinates, if any.
   *
   * @param x The X mouse coordinate.
   * @param y The Y mouse coordinate.
   * @return The node indexes that the nearest edge connects or null.
   */
  protected Point findNearestLink(double x, double y)
  {
    Point edge = null;
    Node tmpnode;
    DPoint3 pos = new DPoint3(), pos2 = new DPoint3();
    DDimension bbox;
    double closest = width_ + height_;
    double xd, yd, dist;

    for (tmpnode = model_.firstNode(); tmpnode != null; tmpnode = model_.nextNode(tmpnode))
    {
      for (int child = tmpnode.firstChild(); child != -1; child = tmpnode.nextChild())
      {
        Node childnode = model_.getNodeFromIndex(child);

        int npoints = 0;
        DPoint3[] path = model_.getEdgePathPoints(tmpnode.getIndex(), child);
        if (path != null)
          npoints = path.length;

        for (int pointindex = 0; pointindex <= npoints; pointindex++)
        {
          if (pointindex == 0)
            pos = tmpnode.getPosition3();
          else
            pos.move(path[pointindex - 1]);

          if (pointindex == npoints)
            pos2 = childnode.getPosition3();
          else
            pos2.move(path[pointindex]);

          pos.transform(viewTransform);
          pos2.transform(viewTransform);

          if ((x >= pos.x - 1 && x <= pos2.x + 1 || x >= pos2.x - 1 &&
               x <= pos.x + 1) && (y >= pos.y - 1 && y <= pos2.y + 1 ||
               y >= pos2.y - 1 && y <= pos.y + 1))
          {
            double dx = pos2.x - pos.x;
            double dy = pos2.y - pos.y;

            if (dx == 0.0 || dy == 0.0)
            {
              dist = 0.0;
            }
            else
            {
              // x distance to line
              xd = Math.abs((dx / dy) * (y - pos.y) + pos.x - x);
              // y distance to line
              yd = Math.abs((dy / dx) * (x - pos.x) + pos.y - y);

              dist = Math.min(xd, yd);
            }

            if (dist < 3.0 && dist < closest)
            {
              closest = dist;
              edge = new Point(tmpnode.getIndex(), child);
            }
          }
        } //for pointindex
      } //for child
    } //for tmpnode
    return edge;
  } //findNearestLink

  /**
   * Search for nearest Composite under given coordinates.
   *
   * @param x_in The X part of the coordinates to look at.
   * @param y_in The Y part of the coordinates to look at.
   * @return The composite if found, null otherwise.
   */
  protected EDU.utwente.CoCompose.lang.Composite findNearestComposite(int x_in,
    int y_in)
  {
    Node thisNode;

    // find nearest node and check if it is a composite
    if ((thisNode = findNearestNode(x_in, y_in)) != null)
    {
      if (thisNode instanceof EDU.utwente.CoCompose.lang.Composite)
      {
        return (EDU.utwente.CoCompose.lang.Composite) thisNode;
      }
    }
    return null;
  } //findNearestComposite

  /**
   * Search for nearest CompositeRole under given coordinates.
   *
   * @param x_in The X part of the coordinates to look at.
   * @param y_in The Y part of the coordinates to look at.
   * @return The composite role if found, null otherwise.
   */
  protected CompositeRole findNearestCompositeRole(int x_in, int y_in)
  {
    Node thisNode;

    // find nearest node and check if it is a composite
    if ((thisNode = findNearestNode(x_in, y_in)) != null)
    {
      if (thisNode instanceof EDU.utwente.CoCompose.lang.CompositeRole)
      {
        return (CompositeRole) thisNode;
      }
    }
    return null;
  } //findNearestCompositeRole

  /**
   * Search for nearest PublishedConcept under given coordinates.
   *
   * @param x_in The X part of the coordinates to look at.
   * @param y_in The Y part of the coordinates to look at.
   * @return The composite role if found, null otherwise.
   */
  protected PublishedConcept findNearestPublishedConcept(int x_in, int y_in)
  {
    Node thisNode;

    // find nearest node and check if it is a composite
    if ((thisNode = findNearestNode(x_in, y_in)) != null)
    {
      if (thisNode instanceof EDU.utwente.CoCompose.lang.PublishedConcept)
      {
        return (PublishedConcept) thisNode;
      }
    }
    return null;
  } //findNearestPublishedConcept

  /**
   * Mouse click dispatcher, that dispatches to handlers depending on the
   * current action being performed with the mouse (e.g. creating an edge).
   *
   * @param e The mouse event.
   */
  protected void dispatchMouseClickAction(MouseEvent e)
  {
    switch (currentMouseAction)
    {
      case MOUSE_NONE: // no current action is in progress
        handleMouseClickNone(e);
        break;
      case MOUSE_EDGE: // an edge is being created
        handleMouseClickLink(e);
        break;
    }
  } //dispatchMouseClickAction

  /**
   * Mouse double click dispatcher, that dispatches to handlers depending on the
   * current action being performed with the mouse (e.g. creating an edge).
   *
   * @param e The mouse event.
   */
  protected void dispatchMouseDoubleClickAction(MouseEvent e)
  {
    switch (currentMouseAction)
    {
      case MOUSE_NONE: // no current action is in progress
        handleMouseDoubleClickNone(e);
        break;
    }
  } //dispatchMouseClickAction

  /**
   * Mouse release dispatcher, that dispatches to handlers depending on the
   * current action being performed with the mouse (e.g. creating an edge).
   *
   * @param e The mouse event.
   */
  protected void dispatchMouseReleaseAction(MouseEvent e)
  {
    switch (currentMouseAction)
    {
      case MOUSE_NODE: // a node is being created
        handleMouseReleaseNode(e);
        break;
      case MOUSE_SELECT: // a selection is being made
        handleMouseReleaseSelect(e);
        break;
      case MOUSE_MSELECT: // a multiple selection is being made
        handleMouseReleaseMSelect(e);
        break;
    }
    setWireframe(false);
  } //dispatchMouseReleaseAction

  /**
   * Mouse drag dispatcher, that dispatches to handlers depending on the
   * current action being performed with the mouse (e.g. creating an edge).
   *
   * @param e The mouse event.
   */
  protected void dispatchMouseDragAction(MouseEvent e)
  {
    switch (currentMouseAction)
    {
      case MOUSE_NODE: // a node is being created
        handleMouseDragNode(e);
        break;
      case MOUSE_EDGE: // an edge is being create
        handleMouseDragLink(e);
        break;
      case MOUSE_SELECT: // a selection is being made
        handleMouseDragSelect(e);
        break;
      case MOUSE_MSELECT: // a multiple selection is being made
        handleMouseDragMSelect(e);
        break;
    }
    // Get transformed position
    DPoint3 pos = new DPoint3(e.getX(), e.getY(), 0.0);
    pos.transform(moveTransform);
    // set new label
    String string = getLabel(pos.x, pos.y, pos.z, true);
    firePropertyChange("Label", "", string);
  } //dispatchMouseDragAction

  /**
   * Mouse right click dispatcher, that dispatches to handlers depending on the
   * current action being performed with the mouse (e.g. creating an edge).
   *
   * @param e The mouse event.
   */
  protected void dispatchMouseRightClick(MouseEvent e)
  {
    switch (currentMouseAction)
    {
      case MOUSE_EDGE: // an edge is being drawn
        deleteSelected();
        break;
      case MOUSE_NONE: // no current action is in progress
        handleMouseRightClickNone(e);
        break;
    }
  } //dispatchMouseClickAction

  /**
   * Mouse click dispatcher, that creates a model node at the given mouse
   * coordinates.
   *
   * @param e The mouse event.
   * @throws CoComposeException if a node could not be created.
   */
  protected void dispatchNodeCreation(MouseEvent e) throws CoComposeException
  {
    int x_in = e.getX();
    int y_in = e.getY();
    // temporary selection mode
    hoverSelect_ = false;
    // dispatch element creation
    int mode = getMouseMode();
    switch (mode)
    {
      case CREATE_CONCEPTS: // new concept
        // if mouse is over a node, select that node instead
        if (findNearestNode(x_in, y_in) != null)
          hoverSelect_ = true;
        else
          model_.setInsertNode(new Concept());
        break;
      case CREATE_SOLROLE: // new solution role
        // if mouse is over a node, select that node instead
        if (findNearestNode(x_in, y_in) != null)
          hoverSelect_ = true;
        else
          model_.setInsertNode(new SolutionRole());
        break;
      case CREATE_COMPOSITES: // new composite
        // if mouse is over a node, select that node instead
        if (findNearestNode(x_in, y_in) != null)
          hoverSelect_ = true;
        else
          model_.setInsertNode(new EDU.utwente.CoCompose.lang.Composite());
        break;
      case CREATE_ROLES:    // add a new role to a composite
        // if mouse is over a compositerole, select that node instead
        if (findNearestCompositeRole(x_in, y_in) != null)
          hoverSelect_ = true;
        else
          createRole(x_in, y_in);
        break;
      case CREATE_PUBLISHED: // add a new published concept to a composite
        // if mouse is over a published concept, select that node instead
        if (findNearestPublishedConcept(x_in, y_in) != null)
          hoverSelect_ = true;
        else
          createPublished(x_in, y_in);
        break;
      case CREATE_REPOS:    // new composite from repository
        // if mouse is over a node, select that node instead
        if (findNearestNode(x_in, y_in) != null)
          hoverSelect_ = true;
        else
          createRepositoryComposite();
        break;
    }
  } //dispatchNodeCreation

  /**
   * Inserts the node that was prepared using dispatchNodeCreation or
   * insert a basic Node.
   *
   * @param x_in The mouse X coordinate.
   * @param y_in The mouse Y coordinate.
   * @see #dispatchNodeCreation
   * @see Node
   */
  protected void insertNode(int x_in, int y_in)
  {
    // save undo state
    updateUndo();
    // Get transformed position
    DPoint3 pos = new DPoint3(x_in, y_in, 0.0);
    pos.transform(moveTransform);
    // unselect all
    if (selectedNode_ != null || selectedLink_ != null)
    {
      unselectItems();
      repaint();
    }
    // insert the element/node
    int index = model_.insertNode();
    // prepare node for moving and fix position and size
    movingNode_ = model_.getNodeFromIndex(index);
    movingNode_.setPosition(pos);
    DDimension bbox = movingNode_.getBoundingBox();
    bbox.width /= scale_;
    bbox.height /= scale_;
    movingNode_.setBoundingBox(bbox.width, bbox.height);
    // if just inserted a composite from repository, retore the roles offset
    if (newReposComposite_ != null)
    {
      newReposComposite_.restoreOffsets();
    }
    // select node
    selectedNode_ = movingNode_;
    movingNode_.setSelected(true);
  } //insertNode

  /**
   * Inserts a new link into the model.
   *
   * @param x_in The mouse X coordinate.
   * @param y_in The mouse Y coordinate.
   * @see Link
   */
  protected void insertLink(int x_in, int y_in)
  {
    // init temp fields
    newLinkNode_ = null;
    lastLinkPoint_ = null;
    // check whether link starts at node
    if ((newLinkNode_ = findNearestNode(x_in, y_in)) != null)
    {
      // save undo state
      updateUndo();
      // unselect all items
      if (selectedNode_ != null || selectedLink_ != null)
      {
        unselectItems();
        // repaint objects in buffer
        Graphics back = backImage_.getGraphics();
        drawObjects(false, back);
        back.dispose();
        // post repaint
        repaint();
      }
      // set temp fields for link creation
      newLinkX_ = x_in;
      newLinkY_ = y_in;
      pathLength_ = 0;
      pathArraySize_ = 10;  // Arbitrary value.
      pathArray_ = new DPoint3[pathArraySize_];
    }
    else
    {
      currentMouseAction = MOUSE_NONE;
    }
  } //insertLink

  /**
   * Makes a selection.
   *
   * @param x_in The mouse X coordinate.
   * @param y_in The mouse Y coordinate.
   * @param multiple If true, adds the new selection to the current selection.
   */
  protected void select(int x_in, int y_in, boolean multiple)
  {
    selected_ = NONE_;

    if (selectedNode_ != null)
    {
      for (Node tmpnode = model_.firstNode(); tmpnode != null && selected_ == NONE_;
        tmpnode = model_.nextNode(tmpnode))
      {
        if (tmpnode.getSelected() && tmpnode.isVisible())
        {
          selectedNode_ = tmpnode;
          DPoint3 posc = tmpnode.getPosition3();
          DDimension bbox = tmpnode.getBoundingBox();
          posc.transform(viewTransform);
          DPoint tr = new DPoint(posc.x + bbox.width / 2.0 * scale_ + 1,
                                 posc.y - bbox.height / 2.0 * scale_ - 1);
          DPoint bl = new DPoint(posc.x - bbox.width / 2.0 * scale_ - 1,
                                 posc.y + bbox.height / 2.0 * scale_ + 1);
          double dist = 16.0, newdist;
          newdist = (posc.x - x_in) * (posc.x - x_in) + (posc.y - y_in) * (posc.y - y_in);
          if (newdist < dist)
          {
            selected_ = CENTER_;
            dist = newdist;
          }
          newdist = (tr.x - x_in) * (tr.x - x_in) + (tr.y - y_in) * (tr.y - y_in);
          if (newdist < dist)
          {
            selected_ = CORNER_;
            if (bbox.height == 0.0 && bbox.width == 0.0)
              selectedRatio_ = 1;
            else
              selectedRatio_ = bbox.height / bbox.width;
            dist = newdist;
          }
          newdist = (posc.x - x_in) * (posc.x - x_in) + (bl.y - y_in) * (bl.y - y_in);
          if (newdist < dist)
          {
            selected_ = BOTTOM_;
            dist = newdist;
          }
          newdist = (bl.x - x_in) * (bl.x - x_in) + (posc.y - y_in) * (posc.y - y_in);
          if (newdist < dist)
            selected_ = LEFT_;
        }
      } //for
      if (selected_ != NONE_)
      {
        DPoint3 tmppos = selectedNode_.getPosition3();
        tmppos.transform(viewTransform);
        movingZ_ = tmppos.z;
        movingX_ = tmppos.x;
        movingY_ = tmppos.y;
        for (Node tmpnode = model_.firstNode(); tmpnode != null; tmpnode = model_.nextNode(tmpnode))
        {
          tmpnode.saveState();
        }
        Enumeration edges = model_.getLinks();
        while (edges.hasMoreElements())
        {
          ((Link) (edges.nextElement())).saveState();
        }
        moveX_ = (double) x_in;
        moveY_ = (double) y_in;
      }
    } //if selectedNode_
    if (selected_ == NONE_)
    {
      if (!multiple)
        unselectItems();
      Node tmpnode = findNearestNode(x_in, y_in);
      Point tmpedge = findNearestLink(x_in, y_in);
      if (tmpnode != null)
      {
        selectedNode_ = tmpnode;
        selectedNode_.setSelected(true);
        selected_ = NONE_;
      }
      else if (tmpedge != null)
      {
        selectedLink_ = tmpedge;
        setLinkSelected_(selectedLink_.x, selectedLink_.y, true);
      }
      else
      {
        currentMouseAction = MOUSE_MSELECT;  // Multiple select.
        multiSelectX_ = x_in;
        multiSelectY_ = y_in;
        multiSelectX2_ = -1;  // First time flag.
      }
    }
    // Get transformed position
    DPoint3 pos = new DPoint3(x_in, y_in, 0.0);
    pos.transform(moveTransform);
    // set new label
    String string = getLabel(pos.x, pos.y, pos.z, true);
    firePropertyChange("Label", "", string);
    // repaint
    repaint();
  } //select

  /**
   * Makes a multiple selection of all elements within the drawn box.
   */
  protected void multiSelect()
  {
    if (multiSelectX2_ == -1)
      return;

    double x1 = (double) Math.min(multiSelectX_, multiSelectX2_);
    double y1 = (double) Math.min(multiSelectY_, multiSelectY2_);
    double x2 = (double) Math.max(multiSelectX_, multiSelectX2_);
    double y2 = (double) Math.max(multiSelectY_, multiSelectY2_);

    Node tmpnode;
    for(tmpnode = model_.firstNode(); tmpnode != null; tmpnode = model_.nextNode(tmpnode))
    {
      DPoint3 pos = tmpnode.getPosition3();
      pos.transform(viewTransform);
      if(pos.x >= x1 && pos.x <= x2 && pos.y >= y1 && pos.y <= y2)
      {
        tmpnode.setSelected(true);
        selectedNode_ = tmpnode;
      }
    }
    repaint();
    fireStateChanged();
  } //multiSelect

  /**
   * Clips a point to a raster of 10 pixels.
   */
  protected DPoint3 clip(DPoint3 pos)
  {
    pos.x = Math.round(pos.x / 10) * 10;
    pos.y = Math.round(pos.y / 10) * 10;
    pos.z = Math.round(pos.z / 10) * 10;
    return pos;
  } //clip

  /**
   * Get bounds of drawing with current transform.
   *
   * @param width The DPoint to store the width in.
   * @param height The DPoint to store the height in.
   */
  protected void getDrawBounds(DPoint width, DPoint height)
  {
    boolean first = true;
    DPoint tmpwidth = new DPoint(), tmpheight = new DPoint();
    DPoint3 pos = new DPoint3();
    Node tmpnode;
    Graphics graphics = getGraphics();
    for (tmpnode = model_.firstNode(); tmpnode != null; tmpnode = model_.nextNode(tmpnode))
    {
      // only consider AbstractConcepts
      if (tmpnode instanceof EDU.utwente.CoCompose.lang.AbstractConcept)
        ((AbstractConcept) tmpnode).getDrawBounds(graphics, viewTransform, tmpwidth, tmpheight);
      else
        continue;

      if (first)
      {
        first = false;
        width.move(tmpwidth);
        height.move(tmpheight);
      }
      else
      {
        width.x = Math.min(width.x, tmpwidth.x);
        width.y = Math.max(width.y, tmpwidth.y);
        height.x = Math.min(height.x, tmpheight.x);
        height.y = Math.max(height.y, tmpheight.y);
      }
      // Check edge path points.
      int tmpnode_index = tmpnode.getIndex();
      for (int child = tmpnode.firstChild(); child != -1; child = tmpnode.nextChild())
      {
        if (tmpnode_index <= child)
        {
          DPoint3[] path = model_.getEdgePathPoints(tmpnode_index, child);
          if (path != null && path.length > 0)
          {
            for (int i = 0; i < path.length; i++)
            {
              pos.move(path[i]);
              pos.transform(viewTransform);
              width.x = Math.min(width.x, pos.x);
              width.y = Math.max(width.y, pos.x);
              height.x = Math.min(height.x, pos.y);
              height.y = Math.max(height.y, pos.y);
            }
          }
        }
      } //for child
    } //for tmpnode
    if (first)
    {
      width.move(0.0, 1.0);
      height.move(0.0, 1.0);
    }
  } //getDrawBounds

  /**
   * Updates the undo state just before a new action will be performed.
   */
  protected void updateUndo()
  {
    try
    {
      Model undoModel = (Model) model_.safeClone();
      undoModels_.push(undoModel);
      while (undoModels_.size() > MAX_UNDO)
      {
        undoModels_.remove(0);
      }
      redoModels_.clear();
    }
    catch (Exception e)
    { // don't save undo snapshot
    }
  } //updateUndo

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Sets whether or not this canvas is enabled. A canvas that is enabled
   * may respond to user input, while a canvas that is not enabled cannot
   * respond to user input.
   *
   * @param enabled If true, canvas is enabled, otherwise disabled.
   */
  public synchronized void setEnabled(boolean enabled)
  {
    // set enabled level
    if (enabled)
    {
      if (enabled_ > 0)
        enabled_--;
    }
    else
      enabled_++;
    // check whether actually enabled or disabled
    // DON'T CALL SUPER METHOD -> cursor setting won't work then
    if (enabled_ == 0)
      setCursor(null);
    else
      setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
  } //setEnabled

  /**
   * Gets the parent frame.
   *
   * @return The parent frame.
   */
  public Frame getFrame()
  {
    return frame_;
  } //getFrame

  /**
   * Gets the horizontal spacing.
   *
   * @return The horizontal spacing.
   */
  public double getHSpacing()
  {
    return 1 / scale_ * hSpacing_;
  } //getHSpacing

  /**
   * Gets the vertical spacing
   */
  public double getVSpacing()
  {
    return 1 / scale_ * vSpacing_;
  }

  /**
   * Gets the selected node.
   *
   * @return The selected node.
   */
  public Node getSelectedNode()
  {
    return selectedNode_;
  } //getSelectedNode

  /**
   * Gets the selected link.
   *
   * @return The selected link (node indexes).
   */
  public Point getSelectedLink()
  {
    return selectedLink_;
  } //getSelectedLink

  /**
   * Selects a node.
   *
   * @param node_id The node id.
   */
  public synchronized void selectNode(int node_id)
  {
    Node node = model_.getNodeFromId(node_id);
    selectedNode_ = node;
    node.setSelected(true);
    repaint();
    fireStateChanged();
  } //selectNode

  /**
   * Gets the viewing scale.
   *
   * @return The viewing scale.
   */
  public double getScale()
  {
    return scale_;
  } //getScale

  /**
   * Sets the viewing scale.
   *
   * @param new_scale The new viewing scale.
   */
  public void setScale(double new_scale)
  {
    // Scale about the center - compute new offset to keep centered
    setOffset(windowSize_.width / 2.0 - (new_scale / scale_) * (windowSize_.width / 2.0 - offset_.x),
              windowSize_.height / 2.0 - (new_scale / scale_) * (windowSize_.height / 2.0 - offset_.y));
    scale_ = new_scale;
    scaleMatrix_.matrix[0][0] = scaleMatrix_.matrix[1][1] =
      scaleMatrix_.matrix[2][2] = scale_;
    updateViewTransform();
    repaint();
    //send event
    fireStateChanged();
  } //setScale

  /**
   * @see #setScale
   */
  public void scale(double scaleval)
  {
    setScale(scaleval);
  } //scale

  /**
   * Selects all nodes.
   */
  public void selectAll()
  {
    for (Node tmpnode = model_.firstNode(); tmpnode != null; tmpnode = model_.nextNode(tmpnode))
    {
      tmpnode.setSelected(true);
    }
    selectedNode_ = model_.firstNode();
    repaint();
    fireStateChanged();
  }

  /**
   * Unselects all graph elements.
   */
  public void unselectItems()
  {
    if (selectedNode_ == null && selectedLink_ == null)
      return;
    for (Node tmpnode = model_.firstNode(); tmpnode != null; tmpnode = model_.nextNode(tmpnode))
      tmpnode.setSelected(false);
    selectedNode_ = null;

    Enumeration edges = model_.getLinks();
    while (edges.hasMoreElements())
    {
      ((Link) (edges.nextElement())).selected = false;
    }
    selectedLink_ = null;
    repaint();
    fireStateChanged();
  } //unselectItems

  /**
   * Deletes all selected graph elements.
   */
  public void deleteSelected()
  {
    if (selectedNode_ != null || selectedLink_ != null)
    {
      updateUndo();
      if(selectedNode_ != null)
      {
        for (Node tmpnode = model_.firstNode(); tmpnode != null; tmpnode = model_.nextNode(tmpnode))
        {
          if (tmpnode.getSelected())
            model_.removeNode(tmpnode);
        }
        selectedNode_ = null;
        computeBounds();
        // since bounds have changed, fire an event
        fireStateChanged();
      }
      if (selectedLink_ != null)
      {
        Enumeration links = model_.getLinks();
        while (links.hasMoreElements())
        {
          Link link = (Link) (links.nextElement());
          if (link.selected)
            model_.removeEdge(link);
        }
      }
      repaint();
      fireStateChanged();
    }
    else if (currentMouseAction == MOUSE_EDGE)  // Drawing edge.
    {
      newLinkNode_ = null;
      currentMouseAction = MOUSE_NONE;
      repaint();
      fireStateChanged();
    }
  } //deleteSelected

  /**
   * Sets wireframe only drawing.
   *
   * @param wireframe If true, draws only wireframe.
   */
  public void setWireframe(boolean wireframe)
  {
    if (wireframe)
      quality_ = 0;
    else
      quality_ = qualityCB_;
  } //setWireframe

  /**
   * Does canvas update.
   *
   * @param adjust_bounds If true, re-computes bounds.
   */
  public synchronized void update(boolean adjust_bounds)
  {
    if (adjust_bounds)
    {
      computeBounds();
      String string = getLabel(0, 0, 0, false);
      firePropertyChange("Label", "", string);
    }
    repaint();
    // post event
    fireStateChanged();
  }

  /**
   * Gets the offset coordinates.
   *
   * @return The offset coordinates.
   */
  public DPoint getOffset()
  {
    DPoint val = new DPoint(0, 0);
    val.x = -(offset_.x - (double) windowSize_.width) - minx_ * scale_;
    val.y = -(offset_.y - (double) windowSize_.height) - miny_ * scale_;
    return val;
  } //getOffset

  /**
   * Sets the offset coordinates.
   *
   * @param xoffset The X offset coordinate.
   * @param yoffset The Y offset coordinate.
   * @param redraw If true, redraws the canvas.
   */
  public void setOffset(double xoffset, double yoffset, boolean redraw)
  {
    offsetx_ = xoffset / scale_;
    offsety_ = yoffset / scale_;
    computeDrawOffset();
    if (redraw)
      repaint();
  } //setOffset

  /**
   * Sets the viewing angles.
   *
   * @param theta The theta viewing angle.
   * @param phi The phi viewing angle.
   */
  public void setViewAngles(double theta, double phi)
  {
    rotxMatrix_.matrix[1][1] = rotxMatrix_.matrix[2][2] = -Math.cos(-phi + Math.PI / 2.0);
    rotxMatrix_.matrix[2][1] = -(rotxMatrix_.matrix[1][2] = -Math.sin(-phi + Math.PI / 2.0));
    rotzMatrix_.matrix[0][0] = rotzMatrix_.matrix[1][1] = Math.cos(-theta);
    rotzMatrix_.matrix[1][0] = -(rotzMatrix_.matrix[0][1] = -Math.sin(-theta));
    updateViewTransform();
    xyPlane_ = (theta == 0.0 && phi == Math.PI / 2.0);
    repaint();
  } //setViewingAngles

  /**
   * Get the position and dimensions of the display window.
   *
   * @return The position and dimensions of the display window.
   */
  public DRect windowRect()
  {
    return new DRect(-offset_.x / scale_, -offset_.y / scale_, windowSize_.width / scale_, windowSize_.height / scale_);
  } //windowRect

  /**
   * Update the display and boundaries, and center the graph in the
   * display window.
   */
  public void center()
  {
    computeBounds();
    setOffset(0.5 * (windowSize_.width - (minx_ + maxx_) * scale_),
              0.5 * (windowSize_.height - (miny_ + maxy_) * scale_));
    repaint();
    // post event
    fireStateChanged();
  } //center

  /**
   * Gets the size of the canvas contents.
   *
   * @return The size of the canvas contents.
   */
  public DDimension getContentSize()
  {
    double w = width_ * scale_ + (double) windowSize_.width * 2.0;
    double h = height_ * scale_ + (double) windowSize_.height * 2.0;
    return new DDimension(w, h);
  } //getContentSize

  /**
   * Changes the class of a selected node, if any. Note that the class of the
   * selected node and the given class name must refer to subclasses of
   * AbstractConcept!
   *
   * @param newClassName New class name.
   * @throws CoComposeException if node class could not be changed.
   */
  public void changeNodeClass(String newClassName)
    throws CoComposeException
  {
    AbstractConcept newConcept;
    Node node = getSelectedNode();

    // throw error if no AbstractConcept selected
    if (!(node instanceof EDU.utwente.CoCompose.lang.AbstractConcept))
      throw new CoComposeException("Nothing selected");
    // create new node
    try
    {
      newConcept = (AbstractConcept) Class.forName(newClassName).newInstance();
      newConcept.copyAttributes((AbstractConcept) node);
    }
    catch (Exception e)
    {
      throw new CoComposeException(e.getMessage());
    }
    changeNode(newConcept);
  } //changeNodeClass

  /**
   * Replaces the selected node, if any.
   *
   * @param newConcept New abtract concept to put in place.
   * @throws CoComposeException if node class could not be changed.
   */
  public void changeNode(AbstractConcept newConcept)
    throws CoComposeException
  {
    Node node = getSelectedNode();

    // throw error if newConcept is null
    if (newConcept == null)
      throw new CoComposeException("New concept may not be null");

    // throw error if no (Role)Concept or Composite is selected
    if (!(node instanceof EDU.utwente.CoCompose.lang.SolutionRole) &&
        !(node instanceof EDU.utwente.CoCompose.lang.Concept)     &&
        !(node instanceof EDU.utwente.CoCompose.lang.Composite))
      throw new CoComposeException("No solution role, concept or composite selected");

    // save undo state
    updateUndo();
    // replace with new node
    model_.replaceNode(node.getIndex(), newConcept);
    newConcept.setPosition(node.getPosition3());
    // if just inserted a composite with roles, retore the roles offset
    if (newConcept instanceof EDU.utwente.CoCompose.lang.Composite)
      ((EDU.utwente.CoCompose.lang.Composite) newConcept).restoreOffsets();
    selectNode(newConcept.getId());
    repaint();
  } //changeNode

  /**
   * Sets the repository container to get composites from.
   *
   * @param repositoryContainer The object that contains the repository tree.
   */
  public void setRepositoryContainer(RepositoryContainer repositoryContainer)
  {
    repository_ = repositoryContainer;
  } //setRepositoryFrame

  /**
   * Sets the behaviour of the mouse.
   *
   * @param mode The new mode (CREATE_NODES | .. | CREATE_ROLES)
   */
  public void setMouseMode(int mode)
  {
    mouseMode_ = mode;
  } //setMouseMode

  /**
   * Gets the behaviour of the mouse.
   *
   * @return mode The new mode (CREATE_NODES | .. | CREATE_ROLES)
   */
  public int getMouseMode()
  {
    return mouseMode_;
  } //setMouseMode

  /**
   * Invokes a properties dialog for concepts, if a Concept object is selected.
   * Returns otherwise.
   */
  public void setNodeProperties()
  {
    Dialog propDialog;
    String classname;
    int dotindex;

    Node which = getSelectedNode();
    if (which == null)
    {
      // don't show dialog
      return;
    }

    // which not null, check class
    classname = which.getClass().getName();
    if ((dotindex = classname.lastIndexOf('.')) != -1)
      classname = classname.substring(dotindex + 1);
    if (which instanceof EDU.utwente.CoCompose.lang.AbstractConcept)
    {
      propDialog = new ConceptPropertiesDialog(frame_,
        "Edit " + classname + " properties", true, (AbstractConcept) which);
    }
    else
    {
      return;
    }
    // show dialog
    propDialog.show();
    // repaint canvas
    repaint();
  } //setNodeProperties

  /**
   * Reports an error.
   *
   * @param error - the error to report
   */
  public void reportError(CoComposeException error)
  {
    AbstractConcept concept = error.getConcept();
    String type = error.getType();
    String message = error.getMessage();
    if (concept != null)
      concept.setFillColor(Color.red);
    MessageDialog errDlg = new MessageDialog(frame_, type, true);
    // switch type
    if (type.equals("Error"))
      errDlg.setIcon(MessageDialog.ICON_ERROR);
    else if (type.equals("Warning"))
      errDlg.setIcon(MessageDialog.ICON_WARNING);
    else if (type.equals("Info"))
      errDlg.setIcon(MessageDialog.ICON_INFO);
    errDlg.setMessage(message);
    errDlg.show();
  } //reportError

  /**
   * Accepts generic notification message.
   *
   * @param sender The message sender.
   */
  public void notify(Object sender)
  {
  } //notify

  /**
   * Gets the background drawing color.
   *
   * @return The background drawing color.
   */
  public Color getBackColor()
  {
    return backColor_;
  } //getBackColor

  /**
   * Sets the background drawing color.
   *
   * @param color The background drawing color.
   */
  public void setBackColor(Color color)
  {
    backColor_ = color;
  } //setBackColor

  /**
   * Gets the line drawing color.
   *
   * @return The line drawing color.
   */
  public Color getLineColor()
  {
    return lineColor_;
  } //getLineColor

  /**
   * Sets the line drawing color.
   *
   * @param color The line drawing color.
   */
  public void setLineColor(Color color)
  {
    lineColor_ = color;
  } //setLineColor

  /**
   * Adds a ChangeListener to the viewport scroller.
   *
   * @param l the listener to be added
   */
  public void addChangeListener(ChangeListener l)
  {
    listenerList.add(ChangeListener.class, l);
  } //addChangeListener

  /**
   * Removes a ChangeListener from the viewport scroller.
   *
   * @param l the listener to be removed
   */
  public void removeChangeListener(ChangeListener l)
  {
    listenerList.remove(ChangeListener.class, l);
  } //removeChangeListener

  /**
   * Gets the model that is being edited in this canvas.
   *
   * @return The model being edited.
   */
  public Model getModel()
  {
    return model_;
  } //getModel

  /**
   * Sets the model that is being edited in this canvas.
   *
   * @param model The model being edited.
   */
  public void setModel(Model model)
  {
    updateUndo();
    model_ = model;
    repaint();
    fireStateChanged();
  } //setModel

  /**
   * Performs undo on last operation.
   */
  public void undo()
  {
    if (canUndo())
    {
      redoModels_.push(model_);
      model_ = (Model) undoModels_.pop();
      repaint();
      fireStateChanged();
    }
  } //undo

  /**
   * Roll back undo on last operation (re-do).
   */
  public void redo()
  {
    if (canRedo())
    {
      undoModels_.push(model_);
      model_ = (Model) redoModels_.pop();
      repaint();
      fireStateChanged();
    }
  } //redo

  /**
   * Checks whether undo is possible.
   *
   * @return True if and only if undo is possible.
   */
  public boolean canUndo()
  {
    return (!undoModels_.isEmpty());
  } //canUndo

  /**
   * Checks whether redo is possible.
   *
   * @return True if and only if redo is possible.
   */
  public boolean canRedo()
  {
    return (!redoModels_.isEmpty());
  } //canRedo

  /**
   * Returns an Encapsulated PostScript representation of the model.
   *
   * @return An Encapsulated PostScript representation of the model.
   */
  public String toEPS()
  {
    Date creationDate = new Date();
    String result = new String();
    int fontsize = (int) Math.ceil((AbstractConcept.FONT_SIZE + 1) * scale_);

    // Get graph drawn boundaries.
    DPoint graphwidth = new DPoint(), graphheight = new DPoint();
    getDrawBounds(graphwidth, graphheight);
    int width = (int) Math.ceil(graphwidth.y - graphwidth.x) + 30;
    int height = (int) Math.ceil(graphheight.y - graphheight.x) + 30;

    // headers
    result += "%!PS-Adobe-3.0 EPSF-3.0\n";
    result += "%%BoundingBox: 0 " + (792 - height) + " " + width + " 792\n" +
              "%%Creator: CoCompose "  + CoCompose.MAJOR_VERSION + "." +
              CoCompose.MINOR_VERSION + "." + CoCompose.MINOR_SUFFIX + "\n" +
              "%%CreationDate: " + creationDate.toString() + "\n" +
              "%%Title: " + frame_.getTitle() + "\n\n";
    // ellipse function
    result += "/ellipse\n{\n" +
              "  gsave\n  newpath\n  /h exch def\n  /w exch def\n" +
              "  translate\n  1 h w div scale\n  0 0 w 2 div 0 360 arc\n" +
              "  1 w h div scale\n  setrgbcolor\n  fill\n" +
              "  1 h w div scale\n  0 0 w 2 div 0 360 arc\n" +
              "  1 w h div scale\n  setrgbcolor\n  stroke\n  grestore\n" +
              "  }\ndef\n\n";
    // rectangle function
    result += "/rectangle\n{\n" +
              "  gsave\n  newpath\n  /h exch def\n  /w exch def\n  /y exch def\n  /x exch def" +
              "  x y moveto\n  w 2 div h 2 div rmoveto\n" +
              "  w neg 0 rlineto\n  0 h neg rlineto\n  w 0 rlineto\n  closepath\n" +
              "  setrgbcolor\n  fill\n" +
              "  x y moveto\n  w 2 div h 2 div rmoveto\n" +
              "  w neg 0 rlineto\n  0 h neg rlineto\n  w 0 rlineto\n  closepath\n" +
              "  setrgbcolor\n  stroke\n  grestore\n" +
              "}\ndef\n\n";
    // label function
    result += "/label\n{\n" +
              "  gsave\n  newpath\n  setrgbcolor\n  /type exch def\n  type 2 eq\n" +
              "  {  /h exch def  /w exch def  } if\n" +
              "  translate\n  1 -1 scale\n  dup length /rows exch def\n" +
              "  type 0 eq\n  { 0 0 moveto } if\n" +
              "  type 1 eq\n  { 0 0 moveto } if\n" +
              "  type 2 eq\n  { 0 rows 2 sub fontsize mul 2 div " +
              "fontsize descent mul add fontsize add moveto } if\n" +
              "  {\n    currentpoint 3 2 roll\n    dup\n" +
              "    stringwidth pop /csw exch def\n" +
              "    /string exch def\n    /y exch def\n    /x exch def\n" +
              "    csw 2 div neg y fontsize sub moveto\n" +
              "    string show\n  } forall\n" +
              "  stroke\n  grestore\n" +
              "}\ndef\n\n";
    // composite type image function
    result += "/modelimage\n{\n" +
              "  gsave\n  /ih exch def\n  /iw exch def\n" +
              "  /imagedataB exch def\n  /imagedataG exch def\n" +
              "  /imagedataR exch def\n  /h exch def\n" +
              "  /w exch def\n  translate\n  w h scale\n" +
              "  iw ih 8 [iw 0 0 ih iw 2 div ih 2 div]" +
              " { imagedataR } { imagedataG } { imagedataB }" +
              " true 3 colorimage\n  grestore\n" +
              "}\ndef\n\n";

    // Output the composite type images.
    EDU.utwente.CoCompose.lang.Composite tmpfeat;
    for (tmpfeat = model_.firstComposite(); tmpfeat != null; tmpfeat = model_.nextComposite(tmpfeat))
    {
      result += tmpfeat.imagePS();
    }
    // model function
    result += "/model\n{\n";
    result += "0 396 translate\n1 1 neg scale\n0 neg 396 neg translate\n\n";
    result += "0.6 setlinewidth\n" +
              "/fontsize " + fontsize + " def\n" +
              "/Helvetica findfont fontsize scalefont setfont\n" +
              "/Helvetica findfont\nbegin\nFontType 1 eq\n" +
              "{ FontBBox }\n{ -2 -2 8 8 } ifelse\nend\n" +
              "/descent 0 def\n";
    result += PSnum_(-graphwidth.x + 20.0) +
              PSnum_(-graphheight.x + 10.0) + "translate\n\n";
    // Output free nodes
    Node tmpnode;
    for (tmpnode = model_.firstFreeNode(); tmpnode != null; tmpnode = model_.nextFreeNode(tmpnode))
    {
      result += tmpnode.toPS(viewTransform);
    }
    // Output the composite images.
    for (tmpfeat = model_.firstComposite(); tmpfeat != null; tmpfeat = model_.nextComposite(tmpfeat))
    {
      result += tmpfeat.toPSimage(viewTransform);
    }
    // Output edges
    Enumeration edges = model_.getLinks();
    while (edges.hasMoreElements())
    {
      Link edge = (Link) (edges.nextElement());
      Node head = edge.head(), tail = edge.tail();
      if (tail.getIndex() <= head.getIndex())
        result += edge.toPS(viewTransform, xyPlane_);
    }
    // Output composite part concepts
    CompositePartConcept tmpconcept;
    for (tmpconcept = model_.firstCompositePartConcept(); tmpconcept != null;
         tmpconcept = model_.nextCompositePartConcept(tmpconcept))
    {
      result += tmpconcept.toPS(viewTransform);
    }
    result += "}\ndef\n\n";
    // run model script and showpage
    result += "model\n";
    return result;
  } //toEPS
} //class ModelCanvas
