/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.lang;

import java.util.Enumeration;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.AttributesImpl;
import EDU.utwente.CoCompose.util.CoComposeException;

/**
 * The role class representing the roles of the CoCompose design language
 * primitive Composite.
 *
 * @author Dennis Wagelaar
 * @version 0.1.15, 20 October 2003
 */

public class CompositeRole extends CompositePartConcept
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  private static final String XMLNAME_ROLE = "CompositeRole";
  private static final String XMLNAME_MULT = "Multiplicity";
  /**
   * Default label position.
   */
  protected static final String DefaultLabelPos = "Below";
  /**
   * Multiplicity value zero (0).
   */
  public static final int MULT_ZERO = 0;
  /**
   * Multiplicity value one (1).
   */
  public static final int MULT_ONE  = 1;
  /**
   * Multiplicity value many (n).
   */
  public static final int MULT_MANY = 2;

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private int multLowerBound_;
  private int multUpperBound_;
  private static final String mult_[] = {"0", "1", "n"};

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a CompositeRole.
   */
  public CompositeRole()
  {
    super();
    // multiplicity text values
    // default multiplicity
    multLowerBound_ = MULT_ZERO;
    multUpperBound_ = MULT_MANY;
  } //CompositeRole

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Create XML from the multiplicity attributes using a SAX DefaultHandler.
   *
   * @param handler The SAX DefaultHandler.
   * @throws SAXException if an XML handler error occurs.
   */
  protected void multToXML(DefaultHandler handler) throws SAXException
  {
    AttributesImpl multAtts = new AttributesImpl();

    multAtts.addAttribute(null, null, "LowerBound", "CDATA",
      mult_[getMultLowerBound()]);
    multAtts.addAttribute(null, null, "UpperBound", "CDATA",
      mult_[getMultUpperBound()]);
    handler.startElement(null, null, XMLNAME_MULT, multAtts);
    handler.endElement(null, null, XMLNAME_MULT);
  } //multToXML

  /**
   * Create XML start from this object using a SAX DefaultHandler, element name
   * and given attributes.
   *
   * @param handler The SAX DefaultHandler.
   * @param element The XML element name.
   * @param atts The XML element attributes.
   * @throws SAXException if an XML handler error occurs.
   */
  protected void startXML(DefaultHandler handler, String element,
    AttributesImpl atts) throws SAXException
  {
    // The method needs to be rewritten from AbstractConcept, since
    // PartOf attribute needs to be suppressed.
    // create new attributes list if none given.
    if (atts == null)
      atts = new AttributesImpl();

    // Add attributes
    atts.addAttribute(null, null, "CompletenessConstraint", "CDATA",
      getCompletenessConstraint());
    atts.addAttribute(null, null, "CorrectnessConstraint", "CDATA",
      getCorrectnessConstraint());
    atts.addAttribute(null, null, "Id", "ID", String.valueOf(getId()));
    atts.addAttribute(null, null, "Name", "CDATA", getName());

    // Write XML
    handler.startElement(null, null, element, atts);

    // GML
    gmlToXML(handler);

    // Comments
    commentsToXML(handler);

    // Specification
    specToXML(handler);

    // Multiplicity
    multToXML(handler);
  } //startXML

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Sets the defaults for this composite role.
   */
  public void setDefaults()
  {
    defaultLabel_ = "Role";
    setShape(DefaultShape);
    setLabelPosition(DefaultLabelPos);
    // set default colors
    setEdgeColor(DEFAULT_EDGE_COLOR);
    setFillColor(DEFAULT_FILL_COLOR);
    setFontColor(DEFAULT_FONT_COLOR);
    setSelectColor(DEFAULT_SELECT_COLOR);
    // default to half Concept size
    setBoundingBox(defaults.getBoundingBox3());
    width_ /= 2;
    height_ /= 2;
  } //setDefaults

  /**
   * Sets the label using the object attributes.
   */
  public synchronized void setLabel()
  {
    super.setLabel();
    boolean featIconOnly = false;
    // check whether parent composite only shows icon
    Composite composite = getComposite();
    if (composite != null)
      featIconOnly = composite.getShowIconOnly();
    if (!featIconOnly)
      setLabel(getLabel() + " [" + mult_[getMultLowerBound()] + ".." +
        mult_[getMultUpperBound()] + "]");
    else
      setLabel("");
  } //setLabel

  /**
   * Gets multiplicity lower bound attribute.
   *
   * @return The multiplicity lower bound attribute.
   */
  public int getMultLowerBound()
  {
    return multLowerBound_;
  } //getMultLowerBound

  /**
   * Sets multiplicity lower bound attribute - updates label as well.<BR>
   * <B>Precondition:</B> MULT_ZERO <= multUpperBound <= MULT_ONE
   *
   * @param multLowerBound_in The new multiplicity lower bound.
   */
  public void setMultLowerBound(int multLowerBound_in)
  {
    // check pre-condition
    if ((multLowerBound_in < MULT_ZERO) || (multLowerBound_in > MULT_ONE))
      return;
    multLowerBound_ = multLowerBound_in;
    // construct label by calling setName
    setLabel();
  } //setMultLowerBound

  /**
   * Gets multiplicity upper bound attribute.
   *
   * @return The multiplicity upper bound attribute.
   */
  public int getMultUpperBound()
  {
    return multUpperBound_;
  } //getMultUpperBound

  /**
   * Sets multiplicity upper bound attribute - updates label as well.<BR>
   * <B>Precondition:</B> MULT_ONE <= multUpperBound <= MULT_MANY
   *
   * @param multUpperBound_in The new multiplicity upper bound.
   */
  public void setMultUpperBound(int multUpperBound_in)
  {
    // check pre-condition
    if ((multUpperBound_in < MULT_ONE) || (multUpperBound_in > MULT_MANY))
      return;
    multUpperBound_ = multUpperBound_in;
    // construct label
    setLabel();
  } //setMultUpperBound

  /**
   * Copies attributes from fromRole to self.
   *
   * @param fromRole The CompositeRole to copy the attributes from.
   */
  public void copyAttributes(CompositeRole fromRole)
  {
    // set inherited attributes
    super.copyAttributes(fromRole);
    // set own attributes
    setMultLowerBound(fromRole.getMultLowerBound());
    setMultUpperBound(fromRole.getMultUpperBound());
  } //copyAttributes

  /**
   * Overwritten scale method to disallow scaling of this object. There's no
   * need to call this method, since it's a stub.
   *
   * @param scalex The X axis scale multiplier.
   * @param scaley The Y axis scale multiplier.
   * @param scalez The Z axis scale multiplier.
   */
  public void scale(double scalex, double scaley, double scalez)
  {
  } //scale

  /**
   * Create XML from this object using a SAX DefaultHandler.
   *
   * @param handler The SAX DefaultHandler.
   * @throws SAXException if an XML handler error occurs.
   */
  public void toXML(DefaultHandler handler) throws SAXException
  {
    // Write XML
    startXML(handler, XMLNAME_ROLE, null);
    endXML(handler, XMLNAME_ROLE);
  } //toXML

  /**
   * Checks the completeness of the composite role using default rules and the
   * specific CompletenessConstraint, if it exists.
   *
   * @throws CoComposeException if the composite role is not completely specified.
   */
  public void checkCompleteness() throws CoComposeException
  {
    // check multiplicity constraints
    if (getMultLowerBound() == MULT_ONE)
    {
      int linkCount = 0;
      Enumeration links = getModel().getLinks(this);
      Object thisLink;
      while (links.hasMoreElements())
      {
        thisLink = links.nextElement();
        linkCount++;
      }
      // there must be at least one link to this role.
      if (linkCount < 1)
        throw new CoComposeException("There must be at least one link to the " +
          getName() + " role", this);
    }
    // inherited check
    super.checkCompleteness();
  } //checkCompleteness

  /**
   * Checks the correctness of the composite role using the default rules and the
   * specific CorrectnessConstraint, if it exists.
   *
   * @throws CoComposeException if the composite is not correctly specified.
   */
  public void checkCorrectness() throws CoComposeException
  {
    // check multiplicity constraints
    if (getMultUpperBound() == MULT_ONE)
    {
      int linkCount = 0;
      Enumeration links = getModel().getLinks(this);
      Object thisLink;
      while (links.hasMoreElements())
      {
        thisLink = links.nextElement();
        linkCount++;
        // there must be at most one link to this role.
        if (linkCount > 1)
          throw new CoComposeException("There can be at most one link to the " +
            getName() + " role", this);
      }
    }
    // inherited check
    super.checkCorrectness();
  } //checkCorrectness

  /**
   * Accepts a ModelVisitor.
   *
   * @param visitor The visitor object to accept.
   * @throws CoComposeException if visiting operation failed.
   */
  public void accept(ModelVisitor visitor) throws CoComposeException
  {
    super.accept(visitor);
    visitor.visitCompositeRole(this);
  } //accept
} //class CompositeRole
