/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.lang;

import java.util.*;
import EDU.utwente.CoCompose.xml.CompositeTypeReader;
import EDU.utwente.CoCompose.util.CoComposeException;
import org.xml.sax.XMLReader;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.AttributesImpl;

/**
 * A Composite generalisation representing a common Composite type that can be
 * stored in the repository.
 *
 * @author Dennis Wagelaar
 * @version 0.1.14, 14 October 2003
 */

public class CompositeType extends Composite
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  // XML parser defaults
  private static boolean setValidation        = false;
  private static boolean setNameSpaces        = true;
  private static boolean setSchemaSupport     = true;
  private static boolean setSchemaFullSupport = false;

  // XML System DTD path relative to application path
  private static final String XMLFILE_SYSTEM = "dtd/CompositeType.dtd";
  // XML names
  private static final String XMLNAME_TYPE   = "CompositeType";

  // Default parser name.
  private static final String
    DEFAULT_PARSER_NAME = "org.apache.xerces.parsers.SAXParser";

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a CompositeType.
   */
  public CompositeType()
  {
    super();
  } //CompositeType()

  /**
   * Constructs a CompositeType from an XML document.
   *
   * @param uri Location of the XML document.
   * @throws Exception if the CompositeType could not be created.
   */
  public CompositeType(String uri) throws Exception
  {
    super();
    // create XML parser
    XMLReader xmlParser = (XMLReader)
      Class.forName(DEFAULT_PARSER_NAME).newInstance();
    xmlParser.setFeature( "http://xml.org/sax/features/validation",
      setValidation);
    xmlParser.setFeature( "http://xml.org/sax/features/namespaces",
      setNameSpaces);
    xmlParser.setFeature( "http://apache.org/xml/features/validation/schema",
      setSchemaSupport );
    xmlParser.setFeature( "http://apache.org/xml/features/validation/schema-full-checking",
      setSchemaFullSupport );
    // hook parser up to reader
    xmlParser.setContentHandler(new CompositeTypeReader(this));
    // start parser
    xmlParser.parse(uri.replace('\\', '/'));
  } //CompositeType(String uri)

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Create XML start from this object using a SAX DefaultHandler, element name
   * and given attributes.
   *
   * @param handler The SAX DefaultHandler.
   * @param element The XML element name.
   * @param atts The XML element attributes.
   * @throws SAXException if an XML handler error occurs.
   */
  protected void startXML(DefaultHandler handler, String element,
    AttributesImpl atts) throws SAXException
  {
    // Whole method needs to be rewritten from AbstractConcept, since
    // Name, Id and PartOf attribute need to be suppressed.
    // create new attributes list if none given.
    if (atts == null)
      atts = new AttributesImpl();

    // Add attributes
    atts.addAttribute(null, null, "Type", "CDATA", getType());
    if (getShowIconOnly())
    {
      atts.addAttribute(null, null, "ShowIconOnly", "CDATA", "Yes");
    }
    else
    {
      atts.addAttribute(null, null, "ShowIconOnly", "CDATA", "No");
    }
    atts.addAttribute(null, null, "CompletenessConstraint", "CDATA",
      getCompletenessConstraint());
    atts.addAttribute(null, null, "CorrectnessConstraint", "CDATA",
      getCorrectnessConstraint());

    // Write XML
    handler.startElement(null, null, element, atts);

    // GML
    gmlToXML(handler);

    // Comments
    commentsToXML(handler);

    // Specification
    specToXML(handler);

    // CompositeRoles
    rolesToXML(handler);

    // PublishedConcepts
    publishedToXML(handler);

    // Solution patterns
    solutionsToXML(handler);

    // Implementation generators
    implementationsToXML(handler);
  } //startXML

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Copies attributes from fromComposite to self.
   *
   * @param fromComposite The Composite to copy the attributes from.
   * @throws CoComposeException if the copy could not be made.
   */
  public void copyAttributes(Composite fromComposite) throws CoComposeException
  {
    PublishedConcept thisConcept;

    // set inherited attributes
    super.copyAttributes(fromComposite);
    // reset name
    setName("");
    // reset PublishedConcepts public names
    Enumeration concepts = getPublishedConcepts();
    while (concepts.hasMoreElements())
    {
      thisConcept = (PublishedConcept) concepts.nextElement();
      thisConcept.setPublicName("");
    }
  } //copyAttributes

  /**
   * Create an XML document from this object using a SAX DefaultHandler.
   *
   * @param handler The SAX DefaultHandler.
   * @param path - the path to the XML file to be written or null.
   * @throws SAXException if an XML handler error occurs.
   */
  public void toXML(DefaultHandler handler, String path) throws SAXException
  {
    String systemId = Model.SystemPath(XMLFILE_SYSTEM, path);

    // write XML document
    handler.startDocument();
    handler.notationDecl(XMLNAME_TYPE, null, systemId);

    startXML(handler, XMLNAME_TYPE, null);
    endXML(handler, XMLNAME_TYPE);

    handler.endDocument();
  } //toXML
} //class CompositeType
