/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002 - 2004
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.lang;

import java.awt.*;
import java.io.File;
import java.util.*;
import EDU.auburn.VGJ.graph.*;
import EDU.auburn.VGJ.util.*;
import EDU.utwente.CoCompose.CoCompose;
import EDU.utwente.CoCompose.xml.ModelReader;
import EDU.utwente.CoCompose.translate.AppliedPatterns;
import EDU.utwente.CoCompose.util.*;
import org.xml.sax.XMLReader;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.AttributesImpl;

/**
 * Represents a design model made in the CoCompose language.
 *
 * @author Dennis Wagelaar
 * @version 0.1.21, 6 April 2004
 */

public class Model extends Graph implements UniqueRealm, Cloneable
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  // XML parser defaults
  private static boolean setValidation        = false;
  private static boolean setNameSpaces        = true;
  private static boolean setSchemaSupport     = true;
  private static boolean setSchemaFullSupport = false;

  // XML System DTD path relative to application path
  private static final String XMLFILE_SYSTEM   = "dtd/DesignModel.dtd";
  // XML names
  private static final String XMLNAME_MODEL    = "DesignModel";
  private static final String XMLNAME_CONCEPTS = "Concepts";
  private static final String XMLNAME_ROLECONC = "SolutionRoles";
  private static final String XMLNAME_COMPOSITES = "Composites";
  private static final String XMLNAME_LINKS    = "ConceptLinks";
  private static final String XMLNAME_LINK     = "ConceptLink";
  private static final String XMLNAME_GML      = "GML";
  private static final String XMLNAME_PRIORITS = "TranslationPriorities";
  private static final String XMLNAME_PRIORITY = "TranslationPriority";
  private static final String XMLNAME_PRI_ITEM = "PriorityItem";

  // Default parser name.
  private static final String
    DEFAULT_PARSER_NAME = "org.apache.xerces.parsers.SAXParser";

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private Node insertNode_;
  private HashSet uniqueIds_ = new HashSet();
  private boolean alignToGrid_ = true;
  private int freeNodeCount_ = 0;
  private String targetLanguage_ = new String();
  private String name_ = new String();
  private Hashtable transPriority_ = new Hashtable();
  private AppliedPatterns appliedPatterns = new AppliedPatterns();

  /**
   * The composite type unique realm.
   */
  protected UniqueRealm typeRealm = new DefaultUniqueRealm();

  /**
   * Collection of contained Links. Format: [Point -> Link].
   * @label ContainsLinks 
   * @supplierCardinality 0..*
   * @link aggregation
   * @associates <{Link}>*/
  private Hashtable links_ = new Hashtable();

  /**
   * Collection of contained AbstractConcepts.
   * @label ContainsConcepts
   * @supplierCardinality 0..*
   * @link aggregation 
   * @associates <{AbstractConcept}>
   */
  private Vector concepts_ = new Vector();

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs an empty Model.
   */
  public Model()
  {
    super();
  } //Model()

  /**
   * Constructs a Model from an XML document.
   *
   * @param uri Location of the XML document.
   * @param name The model name
   * @throws CoComposeException if the Model could not be created.
   */
  public Model(String uri, String name) throws CoComposeException
  {
    this();
    // set name
    setName(name);
    try
    {
      // create XML parser
      XMLReader xmlParser = (XMLReader)
        Class.forName(DEFAULT_PARSER_NAME).newInstance();
      xmlParser.setFeature( "http://xml.org/sax/features/validation",
        setValidation);
      xmlParser.setFeature( "http://xml.org/sax/features/namespaces",
        setNameSpaces);
      xmlParser.setFeature( "http://apache.org/xml/features/validation/schema",
        setSchemaSupport );
      xmlParser.setFeature( "http://apache.org/xml/features/validation/schema-full-checking",
        setSchemaFullSupport );
      // hook parser up to reader
      xmlParser.setContentHandler(new ModelReader(this));
      // start parser
      xmlParser.parse(uri.replace('\\', '/'));
    }
    catch (Exception e)
    {
      throw new CoComposeException(e.getMessage());
    }
  } //Model(String uri)

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /** 
   * Calculates the path to the system DTD.
   * @param systemId - the system DTD relative to application executable path.
   * @return the system DTD relative to the model path.
   */
  protected static String SystemPath(String systemId, String modelPath)
  {
    if (modelPath == null)
      return systemId;
    File appPath = new File("");
    String path = appPath.getAbsolutePath();
    String relativeUri = modelPath.substring(path.length());
    String pathPrefix = new String();
    int separatorIndex = 0;

    // while we find directory separator characters (/ or \), add to system path
    while ((separatorIndex = relativeUri.indexOf(File.separatorChar,
      separatorIndex + 1)) > -1)
    {
      pathPrefix += "../";
    }
    systemId = pathPrefix + systemId;
    return systemId;
  } //dtdPath

  /**
   * Create XML from link using a SAX DefaultHandler.
   *
   * @param handler The SAX DefaultHandler.
   * @param link The link to get the GML data from.
   * @throws SAXException if an XML handler error occurs.
   */
  protected void linkToXML(DefaultHandler handler, Link link)
    throws SAXException
  {
    // create ConceptLink element
    AttributesImpl linkAtts = new AttributesImpl();
    linkAtts.addAttribute(null, null, "HeadConcept", "IDREF",
      String.valueOf(link.head().getId()));
    linkAtts.addAttribute(null, null, "TailConcept", "IDREF",
      String.valueOf(link.tail().getId()));
    handler.startElement(null, null, XMLNAME_LINK, linkAtts);

    // create GML data
    handler.startElement(null, null, XMLNAME_GML, null);
    GMLobject gml = new GMLobject("edge", GMLobject.GMLlist);
    link.setGMLvalues(gml);
    String gmlString = gml.toString(0);
    handler.characters(gmlString.toCharArray(), 0, gmlString.length());
    handler.endElement(null, null, XMLNAME_GML);

    handler.endElement(null, null, XMLNAME_LINK);
  } //linkToXML

  /**
   * Checks whether a new link can be made between these nodes.
   *
   * @param link The link to check.
   * @return True if and only if the new link is legal.
   */
  protected boolean newLinkOk(Link link)
  {
    Node head = link.head();
    Node tail = link.tail();
    Composite composite;
    Node otherNode;
    boolean headIsComposite = false;
    boolean headIsPart = false;
    boolean tailIsComposite = false;
    boolean tailIsPart = false;

    // can't link to self
    if (head.equals(tail))
    {
      return false;
    }
    // head == Composite
    if (!(headIsComposite = head instanceof EDU.utwente.CoCompose.lang.Composite))
    {
      // head == CompositePartConcept
      headIsPart = head instanceof EDU.utwente.CoCompose.lang.CompositePartConcept;
    }
    // tail == Composite
    if (!(tailIsComposite = tail instanceof EDU.utwente.CoCompose.lang.Composite))
    {
      // tail == CompositePartConcept
      tailIsPart = tail instanceof EDU.utwente.CoCompose.lang.CompositePartConcept;
    }
    // composite part concepts can never be linked to their composite or one of
    // the roles in that composite.
    if (headIsPart)
    {
      if (tailIsPart)
      {
        composite = ((CompositePartConcept) tail).getComposite();
        if (composite.hasPartConcept((CompositePartConcept) head))
          return false;
      }
      else if (tailIsComposite)
      {
        composite = (Composite) tail;
        if (composite.hasPartConcept((CompositePartConcept) head))
          return false;
      }
    }
    else if (tailIsPart)
    {
      if (headIsPart)
      {
        composite = ((CompositePartConcept) head).getComposite();
        if (composite.hasPartConcept((CompositePartConcept) tail))
          return false;
      }
      else if (headIsComposite)
      {
        composite = (Composite) head;
        if (composite.hasPartConcept((CompositePartConcept) tail))
          return false;
      }
    }
    // made it through, so node is ok
    return true;
  } //newLinkOk

  /**
   * Enforces correctness constraints for the AbstractConcepts in the link.
   * Removes link if correctness constraints did not apply.
   *
   * @param link The link of which to check the involved AbstractConcepts.
   * @throws CoComposeException if correctness constraints did not apply.
   */
  protected void applyCorrectnessConstraint(Link link) throws CoComposeException
  {
    Node head = link.head();
    Node tail = link.tail();
    // if nodes are AbstractConcepts, check their correctness constraints.
    if (head instanceof EDU.utwente.CoCompose.lang.ConstrainedConcept)
    {
      try
      {
        ((ConstrainedConcept) head).checkCorrectness();
      }
      catch (CoComposeException e)
      {
        // if not ok, restore situation and pass on exception.
        removeEdge(link);
        throw e;
      }
    }
    if (tail instanceof EDU.utwente.CoCompose.lang.ConstrainedConcept)
    {
      try
      {
        ((ConstrainedConcept) tail).checkCorrectness();
      }
      catch (CoComposeException e)
      {
        // if not ok, restore situation and pass on exception.
        removeEdge(link);
        throw e;
      }
    }
  } //applyCorrectnessConstraint

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Links nodes to the given node.
   *
   * @param node The node to link the other nodes to.
   * @param nodes A collection of other nodes to link.
   * @throws CoComposeException if not all links could be made.
   */
  public void linkNodes(Node node, Collection nodes) throws CoComposeException
  {
    Node otherNode;
    Iterator nodesIt = nodes.iterator();

    while (nodesIt.hasNext())
    {
      otherNode = (Node) nodesIt.next();
      insertLink(new Link(node, otherNode, new DPoint3[]{}));
    }
  } //linkNodes

  /**
   * Inserts a link, if it is legal.
   *
   * @param link The link to insert.
   * @throws CoComposeException if the link is illegal.
   */
  public synchronized void insertLink(Link link) throws CoComposeException
  {
    if (newLinkOk(link))
    {
      // ok, so insert
      int n1 = link.tail().getIndex();
      int n2 = link.head().getIndex();
      // remove any old edge
      removeEdge(n1, n2);
      // set new links at the nodes
      link.tail().setChild(n2);
      link.head().setChild(n1);
      // insert this link
      Point index = new Point(n1, n2);
      links_.put(index, link);
      edges_.put(index, link);
      DPoint3[] points = link.points();
      // if not self-link, insert reverse path as well
      if (n1 != n2)
      {
        int length = points.length;
        DPoint3[] reverse_points = new DPoint3[length];
        for (int i = 0; i < length; i++)
          reverse_points[i] = points[length - 1 - i];
        index = new Point(n2, n1);
        Link reverse = new Link(link.head(), link.tail(), reverse_points);
        links_.put(index, reverse);
        edges_.put(index, reverse);
      }
    }
    else // not ok, so report
    {
      throw new CoComposeException("Link from " + link.head().getLabel() +
        " to " + link.tail().getLabel() + " is illegal");
    }
    // now check whether we are within the correctness constraints and repair
    // state to a valid one if needed.
    try
    {
      applyCorrectnessConstraint(link);
    }
    catch (CoComposeException e)
    {
      // not ok, so report
      throw new CoComposeException("Illegal link: " + e.getMessage(), e);
    }
  } //insertLink

  /**
   * Remove the connection from n1 to n2 but leave the nodes in place.
   * @param n1 index of the first node on the edge
   * @param n2 index of the second node on the edge
   */
   public void removeEdge(int n1, int n2)
   {
     links_.remove(new Point(n1, n2));
     links_.remove(new Point(n2, n1));
     super.removeEdge(n1, n2);
   } //removeEdge

  /**
   * Calls saveState for each Node and Link in the model.
   */
  public void saveStates()
  {
    for (Node tmpnode = firstNode(); tmpnode != null;
      tmpnode = nextNode(tmpnode))
    {
      tmpnode.saveState();
    }
    Enumeration links = getLinks();
    while(links.hasMoreElements())
    {
      ((Link) (links.nextElement())).saveState();
    }
  } //saveStates

  /**
   * Gets align to grid attribute.
   *
   * @return The align to grid attribute.
   */
  public boolean getAlignToGrid()
  {
    return alignToGrid_;
  } //getAlignToGrid

  /**
   * Sets align to grid attribute.
   *
   * @param align The align to grid attribute.
   */
  public void setAlignToGrid(boolean align)
  {
    alignToGrid_ = align;
  } //setAlignToGrid

  /**
   * Sets a Node object for insertion in the model. Insertion is performed later
   * using insertNode().
   *
   * @param node_in The Node to insert.
   */
  public void setInsertNode(Node node_in)
  {
    insertNode_ = node_in;
  } //setInsertNode

  /**
   * Insert a new Node object in the model.
   *
   * @return the node index in the internal list.
   */
  public int insertNode()
  {
    // check if a node has already been preset
    if (insertNode_ == null)
    {
      // use default node
      insertNode_ = new Node();
    }
    int index = insertNode(insertNode_);
    // reset node instance preset
    setInsertNode(null);
    return index;
  } //insertNode()

  /**
   * Insert a node in the model.
   *
   * @param node The Node object to insert.
   * @return the node index in the internal list.
   */
  public synchronized int insertNode(Node node)
  {
    // if AbstractConcept and unique realm not set, set to this
    // also set model relation
    if (node instanceof EDU.utwente.CoCompose.lang.AbstractConcept)
    {
      if (((AbstractConcept) node).getUniqueRealm() == null)
        ((AbstractConcept) node).setUniqueRealm(this);
      ((AbstractConcept) node).setModel(this);
      // set composite type unique realm
      if (node instanceof EDU.utwente.CoCompose.lang.Composite)
      {
        ((Composite) node).setTypeUniqueRealm(typeRealm);
      }
      concepts_.add(node);
    }
    // add the node to the model
    if (node instanceof EDU.utwente.CoCompose.lang.CompositePartConcept)
    {
      // add CompositePartConcepts at end of list, so they're drawn last.
      nodeList_.addNodeAt(node, nodeList_.highestIndex() + 1);
    }
    else
    {
      nodeList_.addNode(node);
      freeNodeCount_++;
    }
    Integer nodeId = node.getIdObject();
    if (nodeId == null)
    {
      node.setId(lastTopId_);
      nodeId = node.getIdObject();
    }
    idHash_.put(nodeId, node);
    // Initialize lastTopId_.
    lastTopId_ = 0;
    while(idHash_.containsKey(new Integer(lastTopId_)))
      lastTopId_++;
    // if this is a Composite, add part concepts as well
    if (node instanceof EDU.utwente.CoCompose.lang.Composite)
    {
      Enumeration concepts = ((Composite) node).getPartConcepts();
      while (concepts.hasMoreElements())
      {
        insertNode((CompositePartConcept) concepts.nextElement());
      }
    }
    return node.getIndex();
  } //insertNode(Node node)

  /**
   * Remove the node from the graph (also removes the links connected to it).
   *
   * @param n The node index.
   */
  public synchronized void removeNode(int n)
  {
    // get corresponding node
    Node thisNode = nodeFromIndex(n);
    // if CompositePartConcept, remove from Composite
    if (thisNode instanceof EDU.utwente.CoCompose.lang.CompositePartConcept)
    {
      Composite composite = ((CompositePartConcept) thisNode).getComposite();
      composite.removePartConcept((CompositePartConcept) thisNode);
    }
    // if Composite, removes composite part concepts as well.
    else if (thisNode instanceof EDU.utwente.CoCompose.lang.Composite)
    {
      Enumeration concepts = ((Composite) thisNode).getPartConcepts();
      while (concepts.hasMoreElements())
      {
        removeNode(((CompositePartConcept) concepts.nextElement()));
        // create new Enumeration, since list is altered
        concepts = ((Composite) thisNode).getPartConcepts();
      }
      freeNodeCount_--;
    }
    else
    {
      freeNodeCount_--;
    }
    // unregister identifiers if AbstractConcept, removes model relation
    if (thisNode instanceof EDU.utwente.CoCompose.lang.AbstractConcept)
    {
      ((AbstractConcept) thisNode).unRegister();
      ((AbstractConcept) thisNode).setModel(null);
    }
    // remove this node
    concepts_.remove(thisNode);
    super.removeNode(n);
  } //removeNode

  /**
   * Remove the node from the graph (also removes the links connected to it).
   *
   * @param id The id of the node to remove.
   */
  public void removeNodeFromId(int id)
  {
    // get corresponding node
    Node thisNode = getNodeFromId(id);
    // remove
    removeNode(thisNode);
  } //removeNodeFromId

  /**
   * Replaces the node at position n by the given node, keeping the links.
   *
   * @param n The index of the node to be replaced.
   * @param node The node to be inserted in place.
   * @throws CoComposeException if one of the new links is illegal.
   */
  public synchronized void replaceNode(int n, Node node) throws CoComposeException
  {
    // get corresponding node
    Node thisNode = nodeFromIndex(n);
    // no node at n means no replacing
    if (thisNode == null)
      return;

    // get links
    Collection nodes = getLinkedNodes(thisNode);
    // throw away old node
    removeNode(n);
    // insert new node
    insertNode(node);
    // link links
    linkNodes(node, nodes);
  } //replaceNode

  /**
   * Gets all links in the model.
   *
   * @return All links in the model.
   */
  public Enumeration getLinks()
  {
    Enumeration allLinks = links_.elements();
    Vector links = new Vector();
    Link thisLink;

    // only include one copy of the links, if not directed
    while (allLinks.hasMoreElements())
    {
      thisLink = (Link) allLinks.nextElement();
      if (thisLink.tail().getIndex() <= thisLink.head().getIndex())
        links.add(thisLink);
    }
    return links.elements();
  } //getLinks

  /**
   * Gets all links connected to the given node.
   *
   * @param node The node to which the links should connect.
   * @return an enumeration of links connected to the given node.
   */
  public Enumeration getLinks(Object node)
  {
    Enumeration links = getLinks();
    Vector connectedLinks = new Vector();
    Link thisLink;

    // check all links
    while (links.hasMoreElements())
    {
      thisLink = (Link) links.nextElement();
      // if head or tail is connected to given node, add
      if (thisLink.head().equals(node) || thisLink.tail().equals(node))
      {
        connectedLinks.add(thisLink);
      }
    }
    return connectedLinks.elements();
  } //getLinks

  /**
   * Gets the AbstractConcept with the given name in the given unique realm.
   *
   * @param name The name of the AbstractConcept.
   * @param namespace The namespace of the AbstractConcept name.
   * @return The AbstractConcept if found, null otherwise.
   */
  public AbstractConcept getAbstractConcept(String name, String namespace)
  {
    AbstractConcept concept;
    String cname, realm;

    for (int index = 0; index < concepts_.size(); index++)
    {
      concept = (AbstractConcept) concepts_.get(index);
      cname = concept.getName();
      realm = concept.getUniqueRealm().getName();
      if (cname.equals(name) && realm.equals(namespace))
        return concept;
      if (concept instanceof EDU.utwente.CoCompose.lang.PublishedConcept)
      { // published concepts have a public name by which they can be retrieved
        cname = ((PublishedConcept) concept).getPublicName();
        realm = ((PublishedConcept) concept).getPublicUniqueRealm().getName();
        if (cname.equals(name) && realm.equals(namespace))
          return concept;
      }
    }
    // not found
    return null;
  } //getAbstractConcept(String name, String namespaces)

  /**
   * Gets the AbstractConcepts that were copied from the given AbstractConcept.
   * @param copiedFrom The AbstractConcept that was copied.
   * @return The Collection of AbstractConcepts found.
   */
  public Collection getAbstractConcepts(AbstractConcept copiedFrom)
  {
    AbstractConcept concept;
    Collection concepts = new HashSet();

    // retrieve through copiedFrom field
    for (int index = 0; index < concepts_.size(); index++)
    {
      concept = (AbstractConcept) concepts_.get(index);
      if (concept.isCopiedFrom(copiedFrom))
      {
        concepts.add(concept);
      }
    }
    return concepts;
  } //getAbstractConcepts(AbstractConcept copiedFrom)

  /**
   * Gets the AbstractConcepts that were copied from the given AbstractConcept.
   *
   * @param copiedFrom The AbstractConcept that was copied.
   * @param inRealm Target AbstractConcept should reside within this realm.
   * @return The Collection of AbstractConcepts found.
   */
  public Collection getAbstractConcepts(AbstractConcept copiedFrom,
    UniqueRealm inRealm)
  {
    AbstractConcept concept;
    Collection concepts = new HashSet();

    // retrieve through copiedFrom field
    for (int index = 0; index < concepts_.size(); index++)
    {
      concept = (AbstractConcept) concepts_.get(index);
      if ((concept.isCopiedFrom(copiedFrom)) &&
          (concept.getUniqueRealm() == inRealm))
      {
        concepts.add(concept);
      }
    }
    return concepts;
  } //getAbstractConcepts(AbstractConcept copiedFrom, UniqueRealm inRealm)

  /**
   * Gets the contained AbstractConcepts.
   * @return The collection of AbstractConcepts.
   */
  public Collection getAbstractConcepts()
  {
    return concepts_;
  } //getAbstractConcepts()

  /**
   * Adds a concept form translation priority to the user defined heuristic.
   *
   * @param language The target language in which the concept form is defined.
   * @param form The concept form to add to the priority list.
   */
  public void addTransPriority(String language, String form)
  {
    Vector priorityList = (Vector) transPriority_.get(language);
    if (priorityList == null)
    { // no priority list defined for target language
      priorityList = new Vector();
      transPriority_.put(language, priorityList);
    }
    priorityList.add(form);
  } //addTransPriority

  /**
   * Inserts a concept form translation priority to the user defined heuristic.
   *
   * @param language The target language in which the concept form is defined.
   * @param form The concept form to add to the priority list.
   * @param index The list index to insert the concept form at.
   */
  public void insertTransPriority(String language, String form, int index)
  {
    Vector priorityList = (Vector) transPriority_.get(language);
    if (priorityList == null)
    { // no priority list defined for target language
      priorityList = new Vector();
      transPriority_.put(language, priorityList);
    }
    priorityList.insertElementAt(form, index);
  } //addTransPriority

  /**
   * Removes a concept form translation priority from the user defined heuristic.
   *
   * @param language The target language in which the concept form is defined.
   * @param form The concept form to remove from the priority list.
   */
  public void removeTransPriority(String language, String form)
  {
    Vector priorityList = (Vector) transPriority_.get(language);
    if (priorityList != null)
    { // priority list defined for target language
      priorityList.remove(form);
    }
  } //removeTransPriority

  /**
   * Clears the translation priority user defined heuristics.
   */
  public void clearTransPriorities()
  {
    transPriority_.clear();
  } //clearTransPriority

  /**
   * Gets the concept form translation priorities for the target language.
   *
   * @param language The target language for which to get the priority list.
   * @return The priority list of concept forms.
   */
  public Enumeration getTransPriorities(String language)
  {
    Vector priorityList = null;

    while ((language != null) && (priorityList == null))
    {
      priorityList = (Vector) transPriority_.get(language);
      language = CoCompose.getSuperLanguage(language);
    }
    if (priorityList == null)
    { // no priority list defined for target language
      priorityList = new Vector();
    }
    return priorityList.elements();
  } //getTransPriorities

  /**
   * Gets the list of target languages for which translation priorities have
   * been defined.
   *
   * @return An enumeration of target languages.
   */
  public Enumeration getTransLanguages()
  {
    return transPriority_.keys();
  } //getTransLanguages
  
  /**
   * Copies the translation priorities from another model.
   * @param from - the model to copy from
   */
  public void copyTransPriorities(Model from)
  {
    Enumeration languages = from.getTransLanguages();
    while (languages.hasMoreElements())
    {
      String language = (String) languages.nextElement();
      Enumeration priorities = from.getTransPriorities(language);
      while (priorities.hasMoreElements())
      {
        String priority = (String) priorities.nextElement();
        addTransPriority(language, priority);
      }
    }
  } //copyTransPriorities

  /**
   * Create an XML document from the model using a SAX DefaultHandler.
   *
   * @param handler - the SAX DefaultHandler.
   * @param selectedOnly - if true, only creates XML for selected nodes.
   * @param path - the path to the XML file to be written or null.
   * @throws SAXException if an XML handler error occurs.
   */
  public void toXML(DefaultHandler handler, boolean selectedOnly, String path)
    throws SAXException
  {
    Vector concepts = new Vector();
    Vector roleConcepts = new Vector();
    Vector composites = new Vector();
    Node thisNode;
    Link thisLink;
    String language;
    Enumeration forms;
    String form;
    AttributesImpl priorAtts;
    AttributesImpl atts = new AttributesImpl();
    String systemId = SystemPath(XMLFILE_SYSTEM, path);

    // sort out nodes
    thisNode = firstNode();
    if (selectedOnly)
    { // get first selected node
      while (!thisNode.getSelected())
      {
        thisNode = nextNode(thisNode);
      }
    }
    while (thisNode != null)
    {
      if ((!selectedOnly) || (thisNode.getSelected()))
      {
        if (thisNode instanceof EDU.utwente.CoCompose.lang.Composite)
        {
          composites.add(thisNode);
        }
        else if (thisNode instanceof EDU.utwente.CoCompose.lang.SolutionRole)
        {
          roleConcepts.add(thisNode);
        }
        else if (thisNode instanceof EDU.utwente.CoCompose.lang.Concept)
        {
          concepts.add(thisNode);
        }
      }
      // else don't add
      thisNode = nextNode(thisNode);
    }

    // write XML document
    handler.startDocument();
    handler.notationDecl(XMLNAME_MODEL, null, systemId);
    atts.addAttribute(null, null, "TargetLanguage", "CDATA", getTargetLanguage());
    handler.startElement(null, null, XMLNAME_MODEL, atts);

    // SolutionRoles
    if (roleConcepts.size() > 0)
    {
      handler.startElement(null, null, XMLNAME_ROLECONC, null);
      Enumeration roleConceptsEn = roleConcepts.elements();
      while (roleConceptsEn.hasMoreElements())
      {
        ((SolutionRole) roleConceptsEn.nextElement()).toXML(handler);
      }
      handler.endElement(null, null, XMLNAME_ROLECONC);
    }

    // Concepts
    if (concepts.size() > 0)
    {
      handler.startElement(null, null, XMLNAME_CONCEPTS, null);
      Enumeration conceptsEn = concepts.elements();
      while (conceptsEn.hasMoreElements())
      {
        ((Concept) conceptsEn.nextElement()).toXML(handler);
      }
      handler.endElement(null, null, XMLNAME_CONCEPTS);
    }

    // Composites
    if (composites.size() > 0)
    {
      handler.startElement(null, null, XMLNAME_COMPOSITES, null);
      Enumeration compositesEn = composites.elements();
      while (compositesEn.hasMoreElements())
      {
        ((Composite) compositesEn.nextElement()).toXML(handler);
      }
      handler.endElement(null, null, XMLNAME_COMPOSITES);
    }

    // ConceptLinks
    Enumeration links = getLinks();
    if (links.hasMoreElements())
    {
      handler.startElement(null, null, XMLNAME_LINKS, null);
      while (links.hasMoreElements())
      {
        thisLink = (Link) links.nextElement();
        if (isDirected() || thisLink.tail().getIndex() <= thisLink.head().getIndex())
        {
          if ((!selectedOnly) ||
              (thisLink.head().getSelected() && thisLink.tail().getSelected()))
          { // only save links to selected nodes, if selectedOnly mode
            linkToXML(handler, thisLink);
          }
        }
      }
      handler.endElement(null, null, XMLNAME_LINKS);
    }

    // TranslationPriorities
    Enumeration languages = getTransLanguages();
    if (languages.hasMoreElements())
    {
      handler.startElement(null, null, XMLNAME_PRIORITS, null);
      while (languages.hasMoreElements())
      {
        language = (String) languages.nextElement();
        priorAtts = new AttributesImpl();
        priorAtts.addAttribute(null, null, "TargetLanguage", "CDATA", language);
        handler.startElement(null, null, XMLNAME_PRIORITY, priorAtts);
        forms = getTransPriorities(language);
        while (forms.hasMoreElements())
        {
          form = (String) forms.nextElement();
          priorAtts = new AttributesImpl();
          priorAtts.addAttribute(null, null, "ConceptForm", "CDATA", form);
          handler.startElement(null, null, XMLNAME_PRI_ITEM, priorAtts);
          handler.endElement(null, null, XMLNAME_PRI_ITEM);
        }
        handler.endElement(null, null, XMLNAME_PRIORITY);
      }
      handler.endElement(null, null, XMLNAME_PRIORITS);
    }

    handler.endElement(null, null, XMLNAME_MODEL);
    handler.endDocument();
  } //toXML

  /**
   * Creates a deep copy of the model using only persistent attributes (XML).
   * Also copies name and appliedPatterns properties.
   * @return A deep copy of the model.
   */
  public Object clone()
  {
    Model copy = new Model();
    try
    {
      copy = (Model) safeClone();
    }
    catch (Exception e)
    { // cannot throw any exception because of superclass
      System.err.println("Could not clone model " + getName() + " (" +
        e.getMessage() + ")");
      e.printStackTrace();
    }
    return copy;
  } //clone

  /**
   * Creates a deep copy of the model using only persistent (XML) attributes.
   * Also copies name and appliedPatterns properties. Safe clone operation that
   * throws an exception when cloning fails.
   * @return A deep copy of the model.
   * @throws Exception - if cloning failed.
   * @deprecated
   */
  private Object xmlClone() throws Exception
  {
    Model copy = new Model();
    toXML(new ModelReader(copy), false, null);
    copy.setName(getName());
    copy.setTargetLanguage(getTargetLanguage());
    copy.appliedPatterns = (AppliedPatterns) getAppliedPatterns().clone();
    return copy;
  } //xmlClone

  /**
   * Creates a deep copy of the model.
   * Also copies name and appliedPatterns properties. Safe clone operation that
   * throws an exception when cloning fails.
   * @return A deep copy of the model.
   * @throws Exception - if cloning failed.
   */
  public Object safeClone() throws Exception
  {
    /* This is a terribly long and complex method in dire
     * need of refactoring. I don't consider it worth the
     * effort, however, since CoCompose2 is coming up...
     * If this cloning operation is to be refactored, I'd
     * say it should go in a separate (inner) class.
     */
    Model copy = new Model();
    Hashtable realms_ = new Hashtable();
    HashSet realmNames_ = new HashSet();
    Vector concepts = new Vector();
    Vector roleConcepts = new Vector();
    Vector composites = new Vector();
    Vector newConcepts = new Vector();

    // sort out nodes
    Node thisNode = firstNode();
    while (thisNode != null)
    {
      if (thisNode instanceof EDU.utwente.CoCompose.lang.Composite)
        composites.add(thisNode);
      else if (thisNode instanceof EDU.utwente.CoCompose.lang.SolutionRole)
        roleConcepts.add(thisNode);
      else if (thisNode instanceof EDU.utwente.CoCompose.lang.Concept)
        concepts.add(thisNode);
      thisNode = nextNode(thisNode);
    }

    // SolutionRoles
    Enumeration roleConceptsEn = roleConcepts.elements();
    while (roleConceptsEn.hasMoreElements())
    {
      SolutionRole oldRole = (SolutionRole) roleConceptsEn.nextElement();
      SolutionRole newRole = new SolutionRole();
      newRole.setId(oldRole.getId());
      newRole.copyAttributes(oldRole);
      copy.insertNode(newRole);
    }

    // Concepts
    Enumeration conceptsEn = concepts.elements();
    while (conceptsEn.hasMoreElements())
    {
      Concept oldConcept = (Concept) conceptsEn.nextElement();
      Concept newConcept = new Concept();

      // handle UniqueRealm
      if (oldConcept.getUniqueRealm() != this)
      {
        String from = oldConcept.getUniqueRealm().getName();
        DefaultUniqueRealm realm;
        if (realmNames_.contains(from))
        { // realm already exists
          realm = (DefaultUniqueRealm) realms_.get(from);
        }
        else
        { // new realm needed
          realm = new DefaultUniqueRealm();
          realm.setName(from);
          realmNames_.add(from);
          realms_.put(from, realm);
        }
        newConcept.setUniqueRealm(realm);
      }

      newConcept.setId(oldConcept.getId());
      newConcept.copyAttributes(oldConcept);
      copy.insertNode(newConcept);
      SolutionRole partOf = oldConcept.getPartOf();
      if (partOf != null)
        newConcept.setPartOf((SolutionRole) copy.getNodeFromId(partOf.getId()));
      newConcepts.add(newConcept);
    }

    // Composites
    Enumeration compositesEn = composites.elements();
    while (compositesEn.hasMoreElements())
    {
      Composite oldComposite = (Composite) compositesEn.nextElement();
      Composite newComposite = new Composite();

      // handle UniqueRealm
      if (oldComposite.getUniqueRealm() != this)
      {
        String from = oldComposite.getUniqueRealm().getName();
        DefaultUniqueRealm realm;
        if (realmNames_.contains(from))
        { // realm already exists
          realm = (DefaultUniqueRealm) realms_.get(from);
        }
        else
        { // new realm needed
          realm = new DefaultUniqueRealm();
          realm.setName(from);
          realmNames_.add(from);
          realms_.put(from, realm);
        }
        newComposite.setUniqueRealm(realm);
      }

      newComposite.setId(oldComposite.getId());
      newComposite.copyAttributes(oldComposite);
      // copy CompositePartConcept IDs
      Enumeration parts = newComposite.getPartConcepts();
      while (parts.hasMoreElements())
      {
        CompositePartConcept thisPart = (CompositePartConcept) parts.nextElement();
        thisPart.setId(thisPart.getLastCopiedFrom().getId());
      }
      copy.insertNode(newComposite);
      SolutionRole partOf = oldComposite.getPartOf();
      if (partOf != null)
        newComposite.setPartOf((SolutionRole) copy.getNodeFromId(partOf.getId()));
    }

    // ConceptLinks
    Enumeration links = getLinks();
    if (links.hasMoreElements())
    {
      while (links.hasMoreElements())
      {
        Link thisLink = (Link) links.nextElement();
        if (isDirected() || thisLink.tail().getIndex() <= thisLink.head().getIndex())
        {
          Link newLink = new Link(
            copy.getNodeFromId(thisLink.tail().getId()),
            copy.getNodeFromId(thisLink.head().getId()),
            thisLink);
          copy.insertLink(newLink);
        }
      }
    }

    // Concept implementation constraints
    conceptsEn = newConcepts.elements();
    while (conceptsEn.hasMoreElements())
    {
      Concept newConcept = (Concept) conceptsEn.nextElement();
      Concept oldConcept = (Concept) newConcept.getLastCopiedFrom();
      newConcept.copyConstraints(oldConcept);
    }

    copy.copyTransPriorities(this);    
    copy.setName(getName());
    copy.setTargetLanguage(getTargetLanguage());
    copy.appliedPatterns = (AppliedPatterns) getAppliedPatterns().clone();
    return copy;
  } //safeClone

  /**
   * Checks whether the given value is unique within the realm.
   *
   * @param value The value to check.
   * @return True if and only if the value is unique within the realm.
   */
  public boolean isUnique(String value)
  {
    return (!uniqueIds_.contains(value) && typeRealm.isUnique(value));
  } //isUnique

  /**
   * Registers a new value in the realm.
   *
   * @param value The value to register.
   */
  public void register(String value)
  {
    uniqueIds_.add(value);
  } //register

  /**
   * Unregisters a value from the realm.
   *
   * @param value The value to unregister.
   */
  public void unRegister(String value)
  {
    uniqueIds_.remove(value);
  } //unRegister

  /**
   * Gets the name of the model/unique realm.
   *
   * @return The name of the model/unique realm.
   */
  public String getName()
  {
    return name_;
  } //getName

  /**
   * Sets the name of the model/unique realm.
   *
   * @param name The name of the model/unique realm.
   */
  public void setName(String name)
  {
    name_ = name;
  } //setName

  /**
   * Gets a collection of nodes linked to the given node.
   *
   * @param node The node to find links of.
   * @return A collection of linked nodes.
   */
  public Collection getLinkedNodes(Object node)
  {
    Link thisLink;
    CompositeRole thisRole;
    HashSet nodes = new HashSet();

    Enumeration links = getLinks(node);
    while (links.hasMoreElements())
    {
      thisLink = (Link) links.nextElement();
      if (thisLink.head().equals(node))
      {
        nodes.add(thisLink.tail());
      }
      else if (thisLink.tail().equals(node))
      {
        nodes.add(thisLink.head());
      }
    }
    // if this is a composite, then the role links must be copied as well
    if (node instanceof EDU.utwente.CoCompose.lang.Composite)
    {
      Enumeration roles = ((Composite) node).getRoles();
      while (roles.hasMoreElements())
      {
        thisRole = (CompositeRole) roles.nextElement();
        nodes.addAll(getLinkedNodes(thisRole));
      }
    }
    return nodes;
  } //getLinkedNodes

  /**
   * Gets free nodes count, which is the number of nodes excluding all contained
   * nodes, such as composite roles.
   *
   * @return The number of free nodes.
   */
  public int getFreeNodeCount()
  {
    return freeNodeCount_;
  } //getFreeNodeCount

  /**
   * Gets composite count.
   *
   * @return The number of composites in the model.
   */
  public int getCompositeCount()
  {
    Composite tmpfeat;
    int compositeCount = 0;

    for (tmpfeat = firstComposite(); tmpfeat != null; tmpfeat = nextComposite(tmpfeat))
    {
      compositeCount++;
    }
    return compositeCount;
  } //getCompositeCount

  /**
   * Gets first free, non-contained node from the model.
   *
   * @return The first free node.
   */
  public Node firstFreeNode()
  {
    Node n;

    n = firstNode();
    while (n instanceof EDU.utwente.CoCompose.lang.CompositePartConcept)
    {
      n = nextNode(n);
    }
    return n;
  } //firstFreeNode

  /**
   * Gets next free, non-contained node from the model.
   *
   * @param prev_node The previous node.
   * @return The next free node.
   */
  public Node nextFreeNode(Node prev_node)
  {
    Node n;

    n = nextNode(prev_node);
    while (n instanceof EDU.utwente.CoCompose.lang.CompositePartConcept)
    {
      n = nextNode(n);
    }
    return n;
  } //nextFreeNode

  /**
   * Gets first composite from the model.
   *
   * @return The first composite.
   */
  public Composite firstComposite()
  {
    Node n;

    n = firstNode();
    while ((!(n instanceof EDU.utwente.CoCompose.lang.Composite)) && (n != null))
    {
      n = nextNode(n);
    }
    return ((Composite) n);
  } //firstComposite

  /**
   * Gets next composite from the model.
   *
   * @param prev_node The previous node.
   * @return The next composite.
   */
  public Composite nextComposite(Node prev_node)
  {
    Node n;

    n = nextNode(prev_node);
    while ((!(n instanceof EDU.utwente.CoCompose.lang.Composite)) && (n != null))
    {
      n = nextNode(n);
    }
    return ((Composite) n);
  } //nextComposite

  /**
   * Gets first concept from the model.
   *
   * @return The first concept.
   */
  public Concept firstConcept()
  {
    Node n;

    n = firstNode();
    while ((!(n instanceof EDU.utwente.CoCompose.lang.Concept)) && (n != null))
    {
      n = nextNode(n);
    }
    return ((Concept) n);
  } //firstConcept

  /**
   * Gets next concept from the model.
   *
   * @param prev_node The previous node.
   * @return The next concept.
   */
  public Concept nextConcept(Node prev_node)
  {
    Node n;

    n = nextNode(prev_node);
    while ((!(n instanceof EDU.utwente.CoCompose.lang.Concept)) && (n != null))
    {
      n = nextNode(n);
    }
    return ((Concept) n);
  } //nextConcept

  /**
   * Gets first composite part concept from the model.
   *
   * @return The first composite part concept.
   */
  public CompositePartConcept firstCompositePartConcept()
  {
    Node n;

    n = firstNode();
    while ((!(n instanceof EDU.utwente.CoCompose.lang.CompositePartConcept)) &&
           (n != null))
    {
      n = nextNode(n);
    }
    return ((CompositePartConcept) n);
  } //firstCompositePartConcept

  /**
   * Gets next composite part concept from the model.
   *
   * @param prev_node The previous node.
   * @return The next composite part concept.
   */
  public CompositePartConcept nextCompositePartConcept(Node prev_node)
  {
    Node n;

    n = nextNode(prev_node);
    while ((!(n instanceof EDU.utwente.CoCompose.lang.CompositePartConcept)) &&
           (n != null))
    {
      n = nextNode(n);
    }
    return ((CompositePartConcept) n);
  } //nextCompositePartConcept

  /**
   * Gets target language to translate the model to.
   *
   * @return The target language.
   */
  public String getTargetLanguage()
  {
    return targetLanguage_;
  } //getTargetLanguage

  /**
   * Sets target language to translate the model to.
   *
   * @param language The target language.
   */
  public void setTargetLanguage(String language)
  {
    targetLanguage_ = language;
    appliedPatterns = new AppliedPatterns();
  } //setTargetLanguage

  /**
   * Gets the node at the given position.
   *
   * @param pos The position to look at.
   * @return The node at that position or null.
   */
  public Node getNodeAt(DPoint3 pos)
  {
    Node node;

    for (node = firstNode(); node != null; node = nextNode(node))
    {
      if (node.getPosition3().equals(pos))
      {
        return node;
      }
    }
    return null;
  } //getNodeAt

  /**
   * Gets the number of edges in the model.
   *
   * @return The number of edges.
   */
  public int numberOfEdges()
  {
    return edges_.size();
  } //numberOfEdges

  /**
   * Checks the completeness of the entire model.
   *
   * @throws CoComposeException if a ConstrainedConcept is not completely
   *         specified.
   */
  public void checkCompleteness() throws CoComposeException
  {
    for (Node node = firstNode(); node != null; node = nextNode(node))
    {
      // check each constrained concept
      if (node instanceof EDU.utwente.CoCompose.lang.ConstrainedConcept)
      {
        ((ConstrainedConcept) node).checkCompleteness();
      }
    }
  } //checkCompleteness

  /**
   * Checks the correctness of the entire model.
   *
   * @throws CoComposeException if a ConstrainedConcept is not correctly
   *         specified.
   */
  public void checkCorrectness() throws CoComposeException
  {
    for (Node node = firstNode(); node != null; node = nextNode(node))
    {
      // check each constrained concept
      if (node instanceof EDU.utwente.CoCompose.lang.ConstrainedConcept)
      {
        ((ConstrainedConcept) node).checkCorrectness();
      }
    } //for node
  } //checkCorrecteness

  /**
   * Updates all the composites in the model using the given composite type
   * data.
   *
   * @param ctype The composite type from which to copy the data.
   * @throws CoComposeException if one or more composites could not be
   *         successfully refreshed.
   */
  public void refreshFromComposite(Composite fromComp)
    throws CoComposeException
  {
    String type = fromComp.getType();

    for (Composite comp = firstComposite(); comp != null; comp = nextComposite(comp))
    {
      // check if same type, then copy
      if (comp.getType().equals(type))
      {
        comp.updateAttributes(fromComp);
      }
    } //for comp
  } //refreshFromCompositeType

  /**
   * Gets the applied solution patterns for this model.
   * @return The AppliedPatterns.
   */
  public AppliedPatterns getAppliedPatterns()
  {
    return appliedPatterns;
  } //getAppliedPatterns

  /**
   * Accepts a ModelVisitor.
   *
   * @param visitor The visitor object to accept.
   * @throws CoComposeException if visiting operation failed.
   */
  public void accept(ModelVisitor visitor) throws CoComposeException
  {
    visitor.visitModel(this);
  } //accept
} //class Model
