/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002 - 2004
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Caches applied SolutionPatterns during flattening.
 *
 * @author Dennis Wagelaar
 * @version 0.1.20, 17 March 2004
 */

public class AppliedPatterns implements Cloneable
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Lookup table for the type realm names used for the applied solution patterns
   * Format: String -> UniqueRealm, where the String denotes the file name
   * and the UniqueRealm denotes the type realm.
   */
  private Hashtable appliedPatternTypes = new Hashtable();
  /**
   * Solution pattern model cache table.
   * Format: String -> Model, where the string denotes the file name and the
   * model denotes the solution pattern model.
   */
  private Hashtable appliedSolutionModels = new Hashtable();
  /**
   * Contains the set of filenames of the applied solution patterns to keep
   * track of whether or not a default part should be applied.
   */
  private HashSet appliedPatterns = new HashSet();
  /**
   * Contains filenames of solution pattern that were validated in the
   * determine flattening options phase.
   */
  private HashSet validatedPatterns = new HashSet();

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs an AppliedPatterns.
   */
  public AppliedPatterns()
  {
    super();
  } //AppliedPatterns

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Creates a deep copy of this (cache only).
   * @return The copy.
   */
  public Object clone()
  {
    AppliedPatterns copy = null;
    try
    {
      copy = (AppliedPatterns) super.clone();
      copy.appliedPatternTypes = (Hashtable) appliedPatternTypes.clone();
      copy.appliedSolutionModels = (Hashtable) appliedSolutionModels.clone();
	    copy.appliedPatterns = (HashSet) appliedPatterns.clone();
      // use shared/global copy of validated patterns
    }
    catch (CloneNotSupportedException e)
    {
      e.printStackTrace();
    }
    return copy;
  } //clone

  /**
   * Adds the SolutionPattern and its Model to the cache of applied patterns.
   * @param pattern The applied solution pattern.
   * @param model The applied solution model.
   * @param typeRealm The composite type unique realm.
   */
  public void addAppliedPattern(SolutionPattern pattern, Model model,
    UniqueRealm typeRealm)
  {
    try
    {
      appliedPatternTypes.put(pattern.getFilename(), typeRealm);
      appliedSolutionModels.put(pattern.getFilename(), model);
      appliedPatterns.add(pattern.getFilename());
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
    }
  } //addAppliedPattern

  /**
   * Adds the SolutionPattern as a validated pattern for this run.
   * This information won't be saved on clone().
   * @param pattern the validated solution pattern.
   */
  public void addValidatedPattern(SolutionPattern pattern)
  {
    try
    {
      validatedPatterns.add(pattern.getFilename());
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
    }
  } //addValidatedPattern

  /**
   * Determines whether this pattern is already in the cache.
   * @param filename The solution pattern filename.
   * @return True if the pattern is in the cache.
   */
  public boolean containsPattern(String filename)
  {
    return appliedPatterns.contains(filename);
  } //containsPattern

  /**
   * Retrieves the solution Model for the given solution pattern.
   * @param filename The solution pattern filename.
   * @return The solution Model, if it is in the cache, null otherwise.
   */
  public Model getSolutionModel(String filename)
  {
    return (Model) appliedSolutionModels.get(filename);
  } //getSolutionModel

  /**
   * Retrieves the type realm for the given solution pattern.
   * @param filename The solution pattern filename.
   * @return The type realm, if it is in the cache, null otherwise.
   */
  public UniqueRealm getTypeRealm(String filename)
  {
    return (UniqueRealm) appliedPatternTypes.get(filename);
  } //getTypeRealm

  /**
   * Determines whether the given solution pattern (by filename) has been
   * validated.
   * @param The filename of the SolutionPattern to check.
   * @return True if this solution pattern has been validated.
   */
  public boolean isValidatedPattern(String filename)
  {
    return validatedPatterns.contains(filename);
  } //isValidatedPattern

  /**
   * @see java.lang.Object#toString()
   */
  public String toString()
  {
    String str = "AppliedPatterns{";
    str += "appliedPatterns=" + appliedPatterns.toString();
    str += ",validatedPatterns=" + validatedPatterns.toString();
    str += '}';
    return str;
  }
} //class SolutionPattern
