/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002 - 2004
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Flattens the model using available solution patterns. Uses a depth first
 * strategy, so it can already return after flattening one option.
 *
 * @author Dennis Wagelaar
 * @version 0.1.20, 17 March 2004
 */

public class FlattenVisitorDepthFirst extends ModelVisitor implements Cloneable
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  // If true, does full flatten, otherwise one level flatten.
  private boolean full_ = true;
  // flattened model (use new visitor for other flattening options)
  private Model flatModel_ = null;
  // The hashtable of composites that should not be flattened.
  private HashSet dontFlatten_ = new HashSet();
  // Determines flattening options.
  private FlatteningOptionsVisitor flatteningOptions = new FlatteningOptionsVisitor();
  // Determines the order in which composites should be flattened, based on
  // dependencies.
  private OrderCompositesVisitor order_ = new OrderCompositesVisitor();
  // single linked list to other flattening branches
  private FlattenVisitorDepthFirst nextFlattenVisitor_ = null;
  // Pointer to current flattening option in process branch
  private SolutionPattern currentOption_ = null;
  /**
   * When false, flattening did not change model, so it was not needed.
   */
  protected boolean flattenNeeded = false;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a FlattenVisitor.
   */
  public FlattenVisitorDepthFirst()
  {
  } //FlattenVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Private methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Updates the dontFlatten cache of composites based on the current model's
   * composites.
   * @param currentModel The model currently being flattened
   */
  private void updateDontFlatten(Model currentModel)
  {
    if (log != null)
      log.log("Updating dontFlatten composite cache...");
    HashSet newDontFlatten = new HashSet();
    Iterator dont = dontFlatten_.iterator();
    while (dont.hasNext())
    {
      Composite composite = (Composite) dont.next();
      newDontFlatten.add(
        currentModel.getAbstractConcept(composite.getName(),
          composite.getUniqueRealm().getName()));
    }
    dontFlatten_ = newDontFlatten;
  } //updateDontFlatten

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets full flattening.
   *
   * @return If true, does full flattening, otherwise one level flattening.
   */
  protected boolean getFull()
  {
    return full_;
  } //getFull

  /**
   * Sets the current flattening option for this visitor to continue at.
   * @param currentOption The current flattening option.
   */
  protected void setCurrentOption(SolutionPattern currentOption)
  {
    currentOption_ = currentOption;
  } //setCurrentOption

  /**
   * Gets the current flattening option for this visitor to continue at.
   * @return The current flattening option.
   */
  protected SolutionPattern getCurrentOption()
  {
    return currentOption_;
  } //getCurrentOption

  /**
   * Sets the flattened model so far.
   *
   * @param flatModel The flattened model.
   */
  protected void setFlatModel(Model flatModel)
  {
    flatModel_ = flatModel;
  } //setFlatModel

  /**
   * Checks whether a composite should be flattened based on the set of
   * composites that need not be flattened.
   * @param composite The composite
   * @return True if and only if the composite should be flattened.
   */
  protected boolean dontFlatten(Composite composite)
  {
    return (dontFlatten_.contains(composite) ||
            ( composite.hasImplementations(getFlatModel().getTargetLanguage()) &&
             !composite.hasSolutions() ));
  } //dontFlatten

  /**
   * Adds a composite to the set of composites that should not be flattened.
   * This option is used to prevent infinite loops in flattening.
   *
   * @param composite The composite that should not be flattened.
   */
  protected void addDontFlatten(Composite composite)
  {
    dontFlatten_.add(composite);
  } //addDontFlatten

  /**
   * Create a shared copy of this.
   * @return The copy.
   */
  protected Object clone()
  {
    FlattenVisitorDepthFirst copy = null;
    try
    {
      copy = (FlattenVisitorDepthFirst) super.clone();
      copy.dontFlatten_ = new HashSet(dontFlatten_);
      copy.setFlatModel(null);
      copy.order_ = (OrderCompositesVisitor) order_.clone();
      copy.setCurrentOption(null);
      copy.setNextFlattenVisitor(null);
    }
    catch (CloneNotSupportedException e)
    { // this should never happen
      e.printStackTrace();
    }
    return copy;
  } //clone

  /**
   * Sets the next flattening option branch visitor.
   * @param flatten The next flatten visitor.
   */
  protected void setNextFlattenVisitor(FlattenVisitorDepthFirst nextFlatten)
  {
    if ((nextFlatten != null) && (nextFlattenVisitor_ != null))
    { // link existing visitors to the new one
      nextFlatten.setNextFlattenVisitor(nextFlattenVisitor_);
    }
    nextFlattenVisitor_ = nextFlatten;
  } //setNextFlattenVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets the next flattening option branch visitor.
   * @return The next flatten visitor.
   */
  public FlattenVisitorDepthFirst getNextFlattenVisitor()
  {
    return nextFlattenVisitor_;
  } //getNextFlattenVisitor

  /**
   * Sets the logger object.
   *
   * @param logger The object to log to
   */
  public void setLogger(Logger logger)
  {
    log = logger;
    flatteningOptions.setLogger(logger);
	order_.setLogger(logger);
  } //setLogger

  /**
   * Sets full flattening.
   *
   * @param full If true, does full flattening, otherwise one level flattening.
   */
  public void setFull(boolean full)
  {
    full_ = full;
  } //setFull

  /**
   * Gets the flattened model so far.
   *
   * @return The flattened model.
   */
  public Model getFlatModel()
  {
    return flatModel_;
  } //getFlatModel

  /**
   * Flattens the model.
   *
   * @param model The model to flatten.
   * @throws CoComposeException
   *   if one of the model's constraints are not fulfilled or
   *   if the model could not be flattened.
   */
  public void visitModel(Model model) throws CoComposeException
  {
    Composite composite = null, currentComposite;
    Iterator composites;
    Vector flatterModels = new Vector();
    FlattenVisitorDepthFirst nestedFlatten;

    if (log != null)
    {
      log.log("-- Started flattening for " + model.getName());
      log.incLogLevel();
    }
    try
    {
      // if no current option set, this is the first flattening visitor
      if (getCurrentOption() == null)
      { // enforce model to be correct and complete
  	    model.checkCorrectness();
        model.checkCompleteness();
        // flatten to new model (copy of this)
      	setFlatModel((Model) model.clone());
        if (log != null)
        {
          log.log("Old applied patterns = " + model.getAppliedPatterns().toString());
          log.log("New applied patterns = " + getFlatModel().getAppliedPatterns().toString());
        }
        // determine correct order for composites re. dependencies
        updateDontFlatten(model);
        order_.reset();
        model.accept(order_);
        composites = order_.getComposites().iterator();
      }
      else
      { // if current option exists, go to corresponding composite
        currentComposite = getCurrentOption().getComposite();
        if (log != null)
          log.log("Resuming flattening at next option branch: " +
            currentComposite.getName());
        composites = order_.getComposites().iterator();
        while (composites.hasNext())
        {
          composite = (Composite) composites.next();
          if (currentComposite.getId() == composite.getId())
            break;
        }
        if (composite == null)
        {
          if (log != null)
            log.log("Error: no composites found when resuming flattening");
          throw new CoComposeException(
            "Error: no composites found when resuming flattening",
            currentComposite);
        }
        if (currentComposite.getId() != composite.getId())
        {
          if (log != null)
            log.log("Error: current composite " + currentComposite.getName() +
              " not found when resuming flattening");
          throw new CoComposeException("Error: current composite " +
            currentComposite.getName() + " not found when resuming flattening",
            currentComposite);
        }
        composite.accept(this);
      }
      // flatten each composite
      while (composites.hasNext())
      {
        composite = (Composite) composites.next();
        composite.accept(this);
      }
      if (!flattenNeeded && (log != null))
        log.log("Flattening not needed for " + model.getName());
      // if full flatten requested, flatten all resulting models
      if (getFull() && flattenNeeded)
      {
        // create new flatten visitor for each level
        nestedFlatten = (FlattenVisitorDepthFirst) clone();
        nestedFlatten.flattenNeeded = false; // set to default
        getFlatModel().accept(nestedFlatten);
        setFlatModel(nestedFlatten.getFlatModel());
        if (nestedFlatten.getNextFlattenVisitor() != null)
          setNextFlattenVisitor(nestedFlatten.getNextFlattenVisitor());
      }
    }
    finally
    {
      if (log != null)
      {
        log.decLogLevel();
        log.log("-- Completed flattening for " + model.getName());
      }
    }
  } //visitModel

  /**
   * Flattens a Composite object.
   *
   * @param composite The composite to flatten.
   * @throws CoComposeException
   *   if the composite cannot be flattened for the target language of the model.
   */
  public void visitComposite(Composite composite) throws CoComposeException
  {
    Model model;
    SolutionPattern solution = null;
    Collection flatOptions;
    Iterator options;
    int optionCount = 0;
    FlattenVisitorDepthFirst flatten;
    Stack optionStack = new Stack();

    if (dontFlatten(composite))
      return;
    if (log != null)
    {
      log.log("Flattening " + composite.getName());
      log.incLogLevel();
    }
    try
    {
      // get target model
      model = getFlatModel();
      if (model == null)
      {
        if (log != null)
          log.log("No target model given for flattening " + composite.getName());
        throw new CoComposeException("No target model given for flattening " +
          composite.getName(), composite);
      }
      Composite targetComposite = (Composite) model.getNodeFromId(composite.getId());
      if (targetComposite.isDependent())
      {
        if (log != null)
          log.log(composite.getName() + " depends on published concepts and cannot be flattened yet");
      }
      else
      {
        // if current option exists, go to current option
        if (getCurrentOption() != null)
        {
          solution = getCurrentOption();
          setCurrentOption(null); // reset current option
          if (solution == null)
          {
            if (log != null)
              log.log("Error: current flattening option" +
                " not found when resuming flattening");
            throw new CoComposeException("Error: current flattening option" +
              " not found when resuming flattening");
          }
          // set flag to continue flattening after this
          flattenNeeded = true;
          if (log != null)
            log.log("Resuming at option " + solution.getFilename());
        }
        else
        { // apply flattening options visitor to composite
          flatteningOptions.reset();
          flatteningOptions.setAppliedPatterns(model.getAppliedPatterns());
          flatteningOptions.setTargetLanguage(model.getTargetLanguage());
          composite.accept(flatteningOptions);
          // retrieve flattening options (always returns at least one option)
          flatOptions = flatteningOptions.getOptions();
          if (log != null)
            log.log(flatOptions.size() + " flattening options found for " +
              composite.getName());
          options = flatOptions.iterator();
          while (options.hasNext())
          {
            optionCount++;
            if (log != null)
              log.log("Reading option " + optionCount);
            if (optionCount == 1)
            { // retrieve next option and apply it to the current target model
              solution = (SolutionPattern) options.next();
              if (solution != null)
              { // set flag to continue flattening after this
                flattenNeeded = true;
              }
            }
            else
            { // stack other options
              optionStack.push(options.next());
            }
          } //while options
          while (!optionStack.empty())
          { // create new visitor for each option beyond the first
            flatten = (FlattenVisitorDepthFirst) clone();
            flatten.setFlatModel((Model) model.clone());
            flatten.setCurrentOption((SolutionPattern) optionStack.pop());
            setNextFlattenVisitor(flatten);
          } //while optionStack
        } //if currentSolution
        if (solution != null)
        { // apply solution and remove this composite
          solution.apply(model, log);
        }
        else
        { // tag this composite, so it won't be flattened further (infinite loop)
          addDontFlatten(composite);
        }
      }
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitComposite
} //class FlattenVisitor
