/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.util.*;
import EDU.auburn.VGJ.graph.Node;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Checks the concept form constraints of the model. Supports incremental
 * checking, i.e. only checks for concepts with determined concept forms.
 *
 * @author Dennis Wagelaar
 * @version 0.1.18, 24 November 2003
 */

public class FormConstraintsVisitor extends ModelVisitor
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  // The mapping of concepts to forms: Concept |-> String
  private Hashtable forms_ = new Hashtable();
  // When defined, only checks composites linked to this concept
  private Concept onlyNewConcept_ = null;
  private ConceptFormConstraintsVisitor conceptVisitor_ =
    new ConceptFormConstraintsVisitor(this);
  private CompositeFormConstraintsVisitor compositeVisitor_ =
    new CompositeFormConstraintsVisitor(this);

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a FormConstraintsVisitor.
   */
  public FormConstraintsVisitor()
  {
  } //FormConstraintsVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Retrieves the concept form mapping.
   *
   * @return The concept form mapping.
   */
  protected Hashtable getForms()
  {
    return forms_;
  } //getForms

  /**
   * Gets the OnlyNewConcept attribute. When defined, only checks composites
   * linked to this concept.
   *
   * @return The concept or null.
   */
  protected Concept getOnlyNewConcept()
  {
    return onlyNewConcept_;
  } //getOnlyNewConcept

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Sets the concept forms mapping attribute.
   *
   * @param forms The mapping of concepts to forms: Concept |-> String.
   */
  public void setForms(Hashtable forms)
  {
    forms_ = forms;
  } //setForms

  /**
   * Sets the OnlyNewConcept attribute. When defined, only checks composites
   * linked to this concept.
   *
   * @param concept The concept or null.
   */
  public void setOnlyNewConcept(Concept concept)
  {
    onlyNewConcept_ = concept;
  } //setOnlyNewConcept

  /**
   * Checks the concept form constraints for the model.
   *
   * @param model The model to check.
   * @throws CoComposeException
   *   if one of the model's constraints are not fulfilled or
   *   if the concept form constraints are not met.
   */
  public void visitModel(Model model) throws CoComposeException
  {
    Hashtable forms = getForms();
    Concept onlyNewConcept = getOnlyNewConcept();
    Node node;
    int child;
    Set checkedComposites = new HashSet();
    Iterator concepts;
    AbstractConcept concept;

    if (log != null)
    {
      log.log("-- Started checking concept form constraints for " + model.getName());
      log.incLogLevel();
    }
    try
    {
      // enforce model to be correct and complete
      model.checkCorrectness();
      model.checkCompleteness();
      // check form constraints for each concept
      concepts = model.getAbstractConcepts().iterator();
      while (concepts.hasNext())
      {
        concept = (AbstractConcept) concepts.next();
        // check if concept form determined, otherwise skip check
        if (forms.containsKey(concept))
          concept.accept(conceptVisitor_);
      }
      // see what composites to check
      if (onlyNewConcept != null)
      { // check only composites linked to this concept
        for (child = onlyNewConcept.firstChild(); child != -1;
          child = onlyNewConcept.nextChild())
        {
          node = model.getNodeFromIndex(child);
          if (node instanceof EDU.utwente.CoCompose.lang.CompositeRole)
          {
            concept = ((CompositeRole) node).getComposite();
            if (!checkedComposites.contains(concept))
            {
              concept.accept(compositeVisitor_);
              checkedComposites.add(concept);
            }
          }
        } //for
      }
      else
      { // check form constraints for each composite
        concepts = model.getAbstractConcepts().iterator();
        while (concepts.hasNext())
        {
          concept = (AbstractConcept) concepts.next();
          concept.accept(compositeVisitor_);
        }
      }
    }
    finally
    {
      if (log != null)
      {
        log.decLogLevel();
        log.log("-- Completed checking concept form constraints for " + model.getName());
      }
    }
  } //visitModel

  /**
   * Sets the logger object.
   *
   * @param logger The object to log to
   */
  public void setLogger(Logger logger)
  {
    super.setLogger(logger);
    compositeVisitor_.setLogger(logger);
    conceptVisitor_.setLogger(logger);
  } //setLogger
} //class FormConstraintsVisitor
