/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.util.*;
import EDU.auburn.VGJ.graph.Node;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.CoComposeException;

/**
 * Orders the composites in a list based on the composite interdependencies.
 * If a published concept of one composite fills a role of another compose, for
 * instance, that composite should be flattened first.
 *
 * @author Dennis Wagelaar
 * @version 0.1.11, 6 October 2003
 */

public class OrderCompositesVisitor extends ModelVisitor implements Cloneable
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  // The ordered list of composites
  private Vector composites_ = new Vector();
  // The set of visited composites (for dependency loop detection)
  private Set visited_ = new HashSet();

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a OrderCompositesVisitor.
   */
  public OrderCompositesVisitor()
  {
  } //OrderCompositesVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Adds a composite to the list.
   *
   * @param composite The composite to add.
   */
  protected void addComposite(Composite composite)
  {
    composites_.add(composite);
  } //addComposite

  /**
   * Checks whether the list contains the composite.
   *
   * @param composite The composite to check for.
   * @return True if and only if the list contains the composite.
   */
  protected boolean containsComposite(AbstractConcept composite)
  {
    return composites_.contains(composite);
  } //containsComposite

  /**
   * Checks whether the composite is already being visited.
   *
   * @param The composite to check for.
   * @return True if and only if the composite visit operation has already been
   *              invoked for the given composite.
   */
  protected boolean visited(Composite composite)
  {
    return visited_.contains(composite);
  } //visited

  /**
   * Adds the composite to the visited set.
   *
   * @param The composite to add.
   */
  protected void addVisited(Composite composite)
  {
    visited_.add(composite);
  } //addVisited

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Create a shallow copy of this.
   * @return The copy.
   */
  protected Object clone()
  {
    OrderCompositesVisitor copy = null;
    try
    {
      copy = (OrderCompositesVisitor) super.clone();
      copy.composites_ = new Vector(composites_);
      copy.visited_ = new HashSet(visited_);
    }
    catch (CloneNotSupportedException e)
    { // this should never happen
      e.printStackTrace();
    }
    return copy;
  } //clone

  /**
   * Retrieves the list of composites.
   *
   * @return The ordered list of composites.
   */
  public List getComposites()
  {
    return composites_;
  } //getComposites

  /**
   * Resets the object to the state as it was just after object creation.
   */
  public void reset()
  {
	composites_.clear();
    visited_.clear();
  } //reset

  /**
   * Determines the dependencies of the composites in the model.
   *
   * @param model The model to visit.
   * @throws CoComposeException if a dependency cycle is detected.
   */
  public void visitModel(Model model) throws CoComposeException
  {
    AbstractConcept concept;
    Iterator concepts;

    if (log != null)
    {
      log.log("-- Started determining composite dependencies for " +
        model.getName());
      log.incLogLevel();
    }
    try
    {
      // determine dependencies for each composite
      concepts = model.getAbstractConcepts().iterator();
      while (concepts.hasNext())
      { // check all composites not already in the list
        concept = (AbstractConcept) concepts.next();
        if (!containsComposite(concept))
          concept.accept(this);
      }
    }
    finally
    {
      if (log != null)
      {
        log.decLogLevel();
        log.log("-- Completed determining composite dependencies for " +
          model.getName());
      }
    }
  } //visitModel

  /**
   * Determines the composite dependencies and inserts the composite in the
   * list at the appropriate index.<BR>
   * PRE: The model must be defined for this concept.
   *
   * @param composite The composite to visit.
   * @throws CoComposeException if a dependency cycle is detected.
   */
  public void visitComposite(Composite composite) throws CoComposeException
  {
    Model model = composite.getModel();
    Enumeration roles = composite.getRoles();
    CompositeRole role;
    Node node;
    Composite dependentOn;
    int child;

    if (log != null)
    {
      log.log("Determining dependencies for composite " + composite.getName());
      log.incLogLevel();
    }
    try
    {
      /* Precondition check -- disabled
      // check model
      if (model == null)
      {
        if (log != null)
          log.log("Error: " + concept.getName() + " has no model set");
        throw new CoComposeException(concept.getName() + " has no model set");
      }
      End precondition check */
      if (visited(composite))
      { // check for loop
        if (log != null)
          log.log("Dependency loop detected for " + composite.getName());
        throw new CoComposeException("Dependency loop detected for " +
          composite.getName(), composite);
      }
      addVisited(composite);
      while (roles.hasMoreElements())
      { // see if any roles are filled by published concepts
        role = (CompositeRole) roles.nextElement();
        for (child = role.firstChild(); child != -1; child = role.nextChild())
        { // check all linked nodes
          node = model.getNodeFromIndex(child);
          if (node instanceof EDU.utwente.CoCompose.lang.PublishedConcept)
          { // role filled by published concept, so check that composite first
            dependentOn = ((PublishedConcept) node).getComposite();
            if (!containsComposite(dependentOn))
              dependentOn.accept(this);
          }
        } //for child
      } //while roles
      // all composites on which this composite depends have been added, so add
      // this composite too
      addComposite(composite);
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitComposite
} //class OrderCompositesVisitor
