/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Translates the model using all three translation process steps (flatten,
 * determine concept forms and generate implementation).
 *
 * @author Dennis Wagelaar
 * @version 0.1.15, 20 October 2003
 * @see FlattenVisitor
 * @see DetermineFormsVisitor
 * @see GenerateImplementationVisitor
 */

public class TranslateVisitor extends ModelVisitor
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  private static final String GENIMPL_PATH = "translate.";

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  // transformed models
  private Vector transModels_ = new Vector();

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a TranslateVisitor.
   */
  public TranslateVisitor()
  {
  } //TranslateVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Sets the transformed models so far.
   *
   * @param transModels The collection of transformed models.
   */
  protected void setTransModels(Collection transModels)
  {
    transModels_.clear();
    transModels_.addAll(transModels);
  } //setTransModels

  /**
   * Adds a transformed model to the collection of transformed models.
   *
   * @param transModel The transformed model to add.
   */
  protected void addTransModel(Model transModel)
  {
    transModels_.add(transModel);
  } //addTransModel

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets the transformed models so far.
   *
   * @return The collection of transformed models.
   */
  public Collection getTransModels()
  {
    return transModels_;
  } //getTransModels

  /**
   * Translates the model.
   *
   * @param model The model to translate.
   * @throws CoComposeException
   *   if one of the model's constraints are not fulfilled or
   *   if the model could not be translated.
   */
  public void visitModel(Model model) throws CoComposeException
  {
    FlattenVisitorDepthFirst flatten;
    DetermineFormsVisitor determineForms;
    Iterator models;
    Model currentModel;
    Collection newTransModels = new Vector();
    GenerateImplementationVisitor generateImpl;
    String language = model.getTargetLanguage();

    if (log != null)
    {
      log.log("-- Started translation process for " + model.getName());
      log.incLogLevel();
    }
    try
    {
      // create flatten visitor
      flatten = new FlattenVisitorDepthFirst();
      flatten.setLogger(log);
      flatten.setFull(true);
      // flatten model to new models
      model.accept(flatten);
      currentModel = flatten.getFlatModel();
      // determine forms for each model
      while ((currentModel != null) && (newTransModels.isEmpty()))
      {
        // determineForms operation through visitor construction
        determineForms = new DetermineFormsVisitor();
        determineForms.setLogger(log);
        // try to determine forms to new models
        try
        {
          currentModel.accept(determineForms);
          newTransModels.addAll(determineForms.getResultingModels());
        }
        catch (CoComposeException e)
        { // this model's concept forms could not be determined,
          // just try the next model.
          flatten = flatten.getNextFlattenVisitor();
          if (flatten != null)
          { // flatten next option
            model.accept(flatten);
            currentModel = flatten.getFlatModel();
          }
          else
          { // no options left
            currentModel = null;
          }
        }
      } //while
      setTransModels(newTransModels);
      // test how many models we have
      if (newTransModels.size() == 0)
      { // No models! This means translation fails
        if (log != null)
          log.log("Could not translate " + model.getName() +
            ": No design alternatives left after flattening and determining concept forms");
        throw new CoComposeException("Could not translate " + model.getName() +
            ": No design alternatives left after flattening and determining concept forms");
      }
      else if (newTransModels.size() > 1)
      { // More than one resulting model, user must choose alternative
        if (log != null)
          log.log("More than one model resulted from flattening and determining concept forms" +
            ": Please select a resulting model manually and then generate implementation");
        throw new CoComposeException("More than one model resulted from flattening and determining concept forms" +
            ": Please select a resulting model manually and then generate implementation",
            null, "Info");
      }
      // We now have exactly one model left, so generate implementation
      try
      { // dynamically load implementation visitor using target language
        generateImpl = (GenerateImplementationVisitor) Class.forName(
          GENIMPL_PATH + language + "ImplementationVisitor").newInstance();
        generateImpl.setLogger(log);
      }
      catch (Exception exc)
      { // external reason caused generating implementation to fail
        // (e.g. visitor class not found)
        if (log != null)
          log.log("Generating implementation failed: " + exc.getMessage());
        throw new CoComposeException("Generating implementation failed: " +
          exc.getMessage());
      }
      // generate implementation
      models = newTransModels.iterator();
      currentModel = (Model) models.next();
      currentModel.accept(generateImpl);
      // if generating implementation succeeded, we don't need the model
      models.remove();
      setTransModels(newTransModels);
    }
    finally
    {
      if (log != null)
      {
        log.decLogLevel();
        log.log("-- Completed translation process for " + model.getName());
      }
    }
  } //visitModel
} //class TranslateVisitor
