/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.util;

/**
 * Represents a valid programming language identifier.
 *
 * @author Dennis Wagelaar
 * @version 0.1.06, 7 February 2003
 */

public class Identifier extends UniqueString
{
  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs an Identifier.
   *
   * @param realm The realm in which the value has to be unique.
   * @throws CoComposeException if the new identifier ("") is not unique or valid.
   */
  public Identifier(UniqueRealm realm) throws CoComposeException
  {
    this(realm, "");
  } //Identifier()

  /**
   * Constructs a Identifier.
   *
   * @param realm The realm in which the value has to be unique.
   * @param value The value to set this string to.
   * @throws CoComposeException if the new identifier is not unique or valid.
   */
  public Identifier(UniqueRealm realm, String value) throws CoComposeException
  {
    super(realm, value);
  } //Identifier(String value)

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Check whether the string value is a valid identifier.
   *
   * @param value The string value.
   * @return True if and only if the string value is valid.
   */
  protected boolean valueOk(String value)
  {
    char c;

    // An identifier must be at least one character long
    if (value.length() < 1)
      return false;

    // check first character
    c = value.charAt(0);
    if ( ((c < 'A') || (c > 'Z')) &&
         ((c < 'a') || (c > 'z')) )
    {
      return false;
    }
    // check following characters
    for (int i = 1; i < value.length(); i++)
    {
      c = value.charAt(i);
      if ( ((c < 'A') || (c > 'Z')) &&
           ((c < 'a') || (c > 'z')) &&
           ((c < '0') || (c > '9')) &&
           (c != '_') && (c != '[') &&
           (c != ']') )
      {
        return false;
      }
    }
    // made it through
    return true;
  } //valueOk

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Sets the string value.
   *
   * @param value The new string value.
   * @throws CoComposeException if the new value is not unique or valid.
   */
  public void setValue(String value) throws CoComposeException
  {
    if (valueOk(value))
    {
      super.setValue(value);
    }
    else
    {
      throw new CoComposeException("\"" + value + "\" is an invalid identifier");
    }
  } //setValue
} //class Identifier
