/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.util;

import java.util.*;
import EDU.auburn.VGJ.graph.*;
import EDU.utwente.CoCompose.gui.ModelUpdate;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.CoComposeException;

/**
 * Positions all entities of a model using a source model.
 *
 * @author Dennis Wagelaar
 * @version 0.1.06, 7 March 2003
 */

public class PositionVisitor extends ModelVisitor
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private Model sourceModel_;
  // cached copy of the model being visited
  private Model thisModel_ = null;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a PositionVisitor.
   *
   * @param source The source model to read the positions from.
   */
  public PositionVisitor(Model source)
  {
    sourceModel_ = source;
  } //PositionVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets the source model for positioning.
   *
   * @return The source model.
   */
  protected Model getSourceModel()
  {
    return sourceModel_;
  } //getSourceModel

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Iterates over all free nodes that are also AbstractConcepts.
   *
   * @param model The model to iterate over.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitModel(Model model) throws CoComposeException
  {
    Node node;
    Link link;

    // cache model copy
    thisModel_ = model;
    // position nodes
    for (node = model.firstFreeNode(); node != null; node = model.nextFreeNode(node))
    {
      if (node instanceof EDU.utwente.CoCompose.lang.AbstractConcept)
      {
        if (node instanceof EDU.utwente.CoCompose.lang.Composite)
          ((Composite) node).saveOffsets();
        ((AbstractConcept) node).accept(this);
      }
    } //for node
    // position links (intermediate points)
    Enumeration links = model.getLinks();
    while (links.hasMoreElements())
    {
      link = (Link) links.nextElement();
      link.accept(this);
    } //while links
  } //visitModel

  /**
   * Positions an AbstractConcept object using the source model.<BR>
   * PRE: The model attribute must be defined for the concept.
   *
   * @param concept The abstract concept to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitAbstractConcept(AbstractConcept concept) throws CoComposeException
  {
    Model srcModel = getSourceModel();
    Model model = concept.getModel();
    AbstractConcept source;
    String namespace = concept.getUniqueRealm().getName();
    ModelUpdate display = getUpdate();

    /* Precondition check -- disabled
    // check model
    if (model == null)
    {
      throw new CoComposeException(concept.getName() + " has no model set");
    }
    End precondition check */
    // check namespace
    if (namespace.equals(model.getName()))
    { // use source model name instead of the concept's model name
      namespace = srcModel.getName();
    }
    // retrieve source concept
    source = srcModel.getAbstractConcept(concept.getName(), namespace);
    if (source == null)
    { // try last part of namespace for backward compatibility
      namespace = namespace.substring(namespace.lastIndexOf('_') + 1);
      source = srcModel.getAbstractConcept(concept.getName(), namespace);
    }
    if (source == null)
    { // try retrieval by ID
      source = (AbstractConcept) srcModel.getNodeFromId(concept.getId());
    }
    if (source != null)
    {
      concept.setPosition(source.getPosition3());
      if (display != null)
        display.update(false);
    }
  } //visitAbstractConcept

  /**
   * Positions a Composite object using the source model.
   *
   * @param composite The composite to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitComposite(Composite composite) throws CoComposeException
  {
    ModelUpdate display = getUpdate();

    composite.restoreOffsets();
    if (display != null)
      display.update(false);
  } //visitComposite

  /**
   * Positions a Link object intermediate points using the source model.<BR>
   *
   * @param link The link to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitLink(Link link) throws CoComposeException
  {
    Edge edge;
    Model srcModel = getSourceModel();
    Node tail = link.tail();
    Node head = link.head();
    Node tailSrc = srcModel.getNodeAt(tail.getPosition3());
    Node headSrc = srcModel.getNodeAt(head.getPosition3());
    ModelUpdate display = getUpdate();

    if ((headSrc != null) && (tailSrc != null))
    { // head and tail nodes exist in source model
      edge = srcModel.getEdge(tailSrc.getIndex(), headSrc.getIndex());
      if (edge != null)
      { // link equivalent exists in source model
        if (edge.points().length > 0)
        {
          link = new Link(tail, head, edge);
          // re-insert if possible
          if (thisModel_ != null)
          {
            thisModel_.insertLink(link);
            if (display != null)
              display.update(false);
          } //if thisModel
        } //if edge.points
      } //if edge
    } //if headSrc
  } //visitLink
} //class PositionVisitor
