/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.util;

/**
 * A string that is guaranteed unique or throws an exception.
 *
 * @author Dennis Wagelaar
 * @version 0.0.23, 20 June 2002
 */

public class UniqueString
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  /* The internal String object. */
  private String value_;
  /* Uniqueness realm in which the value should be unique. */
  private UniqueRealm realm_;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a UniqueString.
   *
   * @param realm The realm in which the value has to be unique.
   * @throws CoComposeException if the new UniqueString ("") is not unique.
   */
  public UniqueString(UniqueRealm realm) throws CoComposeException
  {
    this(realm, "");
  } //UniqueString()

  /**
   * Constructs a UniqueString.
   *
   * @param realm The realm in which the value has to be unique.
   * @param value The value to set this string to.
   * @throws CoComposeException if the new UniqueString is not unique.
   */
  public UniqueString(UniqueRealm realm, String value) throws CoComposeException
  {
    realm_ = realm;
    setValue(value);
  } //UniqueString(String value)

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets the string value.
   *
   * @return The string value.
   */
  public String getValue()
  {
    return value_;
  } //getValue

  /**
   * Sets the string value.
   *
   * @param value The new string value.
   * @throws CoComposeException if the new value is not unique.
   */
  public void setValue(String value) throws CoComposeException
  {
    if (realm_ != null)
    {
      realm_.unRegister(value_);
      if (realm_.isUnique(value))
      {
        value_ = value;
        realm_.register(value);
      }
      else
      {
        realm_.register(value_);
        throw new CoComposeException("String \"" + value +
          "\" not unique in realm " + realm_.getName());
      }
    }
    else // no realm is always unique
    {
      value_ = value;
    }
  } //setValue
} //class UniqueString
